import itertools

toSwizzle = ["xy", "xyz", "xyzw"]

for swiz in toSwizzle:
    for perm in itertools.permutations(list(swiz)):
        term = ", ".join(perm)
        print(f"""inline Vec<DTYPE, {len(perm)}> {"".join(perm)}() const {{ return {{ {term} }}; }}""")



from argumentHelper import *

# Detect LibRapid features
features = []
try:
	with open("../../configuration.txt", "r") as file:
		for line in file:
			if len(line) == 0 or line.startswith("#"):
				continue
			
			print("FEATURE:", line.strip())
			try:
				args = line.split()
				features.append((args[0], args[2])) # Cut out the "="
			except:
				pass
except:
	pass

resStr = ""

types = [
	# Scalar types
	"int64_t",
	"double",
	"librapid::mpz",
	"librapid::mpf",
	"librapid::mpq",
	"librapid::mpfr",

	# Complex number types
	"librapid::Complex<float>",
	"librapid::Complex<double>",
	"librapid::Complex<librapid::mpfr>"
]

functions = []

for type in types:
	t = type
	typename = "{}".format(t)
	constRef = "const {} &".format(t)
	ref = "{} &".format(t)
	int64_t = "int64_t"
	extent = "librapid::Extent"
	stdString = "std::string"
	stdStringConstRef = "const std::string &"

	if type not in ["librapid::mpz", "librapid::mpq", "librapid::mpf"]:
		functions += [
			Function("abs", [Argument(constRef, "val")], "return lrc::abs(val);"),
			Function("floor", [Argument(constRef, "val")], "return lrc::floor(val);"),
			Function("ceil", [Argument(constRef, "val")], "return lrc::ceil(val);"),
			Function("pow", [Argument(constRef, "base"), Argument(constRef, "power")], "return lrc::pow(base, power);"),
			Function("sqrt", [Argument(constRef, "val")], "return lrc::sqrt(val);"),
			Function("exp", [Argument(constRef, "val")], "return lrc::exp(val);"),
			Function("exp2", [Argument(constRef, "val")], "return lrc::exp2(val);"),
			Function("exp10", [Argument(constRef, "val")], "return lrc::exp10(val);"),
			Function("log", [Argument(constRef, "val")], "return lrc::log(val);"),
			Function("log2", [Argument(constRef, "val")], "return lrc::log2(val);"),
			Function("log10", [Argument(constRef, "val")], "return lrc::log10(val);"),

			Function("sin", [Argument(constRef, "val")], "return lrc::sin(val);"),
			Function("cos", [Argument(constRef, "val")], "return lrc::cos(val);"),
			Function("tan", [Argument(constRef, "val")], "return lrc::tan(val);"),

			Function("asin", [Argument(constRef, "val")], "return lrc::asin(val);"),
			Function("acos", [Argument(constRef, "val")], "return lrc::acos(val);"),
			Function("atan", [Argument(constRef, "val")], "return lrc::atan(val);"),
			Function("atan2", [Argument(constRef, "a"), Argument(constRef, "b")], "return lrc::atan2(a, b);"),

			Function("csc", [Argument(constRef, "val")], "return lrc::csc(val);"),
			Function("sec", [Argument(constRef, "val")], "return lrc::sec(val);"),
			Function("cot", [Argument(constRef, "val")], "return lrc::cot(val);"),

			Function("acsc", [Argument(constRef, "val")], "return lrc::acsc(val);"),
			Function("asec", [Argument(constRef, "val")], "return lrc::asec(val);"),
			Function("acot", [Argument(constRef, "val")], "return lrc::acot(val);"),

			Function("sinh", [Argument(constRef, "val")], "return lrc::sinh(val);"),
			Function("cosh", [Argument(constRef, "val")], "return lrc::cosh(val);"),
			Function("tanh", [Argument(constRef, "val")], "return lrc::tanh(val);"),

			Function("asinh", [Argument(constRef, "val")], "return lrc::asinh(val);"),
			Function("acosh", [Argument(constRef, "val")], "return lrc::acosh(val);"),
			Function("atanh", [Argument(constRef, "val")], "return lrc::atanh(val);"),

			Function("mod", [Argument(constRef, "val"), Argument(constRef, "divisor")], "return lrc::mod(val, divisor);"),
			Function("round", [Argument(constRef, "val"), Argument("int64_t", "dp", "0")], "return lrc::round(val, dp);"),
			Function("roundSigFig", [Argument(constRef, "val"), Argument("int64_t", "dp", "3")], "return lrc::roundSigFig(val, dp);"),
			Function("roundTo", [Argument(constRef, "val"), Argument(constRef, "num")], "return lrc::roundTo(val, num);"),
			Function("roundUpTo", [Argument(constRef, "val"), Argument(constRef, "num")], "return lrc::roundUpTo(val, num);"),
		]

functions += [
	Function("map", [Argument("double", "val"), 
					 Argument("double", "start1"),
					 Argument("double", "stop1"),
					 Argument("double", "start2"),
					 Argument("double", "stop2")], "return lrc::map(val, start1, stop1, start2, stop2);"),
	Function("random", [Argument("double", "lower", "0"), Argument("double", "upper", "1"), Argument("int64_t", "seed", "-1")], "return librapid::random(lower, upper, seed);"),
	Function("randint", [Argument("int64_t", "lower", "0"), Argument("int64_t", "upper", "0"), Argument("int64_t", "seed", "-1")], "return librapid::randint(lower, upper, seed);"),
	Function("trueRandomEntropy", [], "return librapid::trueRandomEntropy();"),
	Function("trueRandom", [Argument("double", "lower", "0"), Argument("double", "upper", "1")], "return librapid::trueRandom(lower, upper);"),
	Function("trueRandint", [Argument("int64_t", "lower", "0"), Argument("int64_t", "upper", "1")], "return librapid::trueRandint(lower, upper);"),
	Function("randomGaussian", [], "return librapid::randomGaussian();"),
	Function("pow10", [Argument("int64_t", "exponent")], "return librapid::pow10(exponent);"),
	Function("lerp", [Argument("double", "a"), Argument("double", "b"), Argument("double", "t")], "return librapid::lerp(a, b, t);"),
]

for i in range(len(functions)):
	function = functions[i]
	if isinstance(function,Function):
		resStr += "\tmodule" + function.gen()
	else:
		resStr += "\t" + function
	resStr += ";\n"

def write(path:str):
	with open(path.format(type), "w") as file:
		file.write("""
#include <librapid/librapid.hpp>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/functional.h>
#include <functional>
#include <string>

// Just remove these. They're pointless
#ifdef min
#undef min
#endif

#ifdef max
#undef max
#endif

namespace lrc = librapid;
namespace py = pybind11;

""")

		file.write("void init_math(py::module &module) {\n")
		file.write(resStr)
		file.write("\n}")

if __name__ == "__main__":
	write("../autogen/mathInterface.cpp")

	print("\"${CMAKE_CURRENT_SOURCE_DIR}/src/librapid/python/autogen/mathInterface.cpp\"")
	print("\n")
	print("void init_math(py::module &);")
	print("\n")
	print("init_math(module);")


class Argument:
	def __init__(self, type:str, name:str, default:str = None):
		self.type = type.strip().lstrip()
		self.name = name.strip().lstrip()
		self.default = default

	def setType(self, type:str):
		self.type = type

	def hasDefault(self):
		return self.default is not None

	def __str__(self) -> str:
		return self.type + " " + self.name

class Function:
	def __init__(self, name:str, args:list, op:str, **kwargs):
		self.name = name
		self.args = args
		self.op = op
		self.static = False

		if "static" in kwargs:
			self.static = kwargs["static"]

	def gen(self):
		inputArgs = ""
		if len(self.args) != 0:
			for i in range(len(self.args)):
				inputArgs += "{0} {1}".format(self.args[i].type, self.args[i].name)

				if i + 1 < len(self.args):
					inputArgs += ", "
		
		start = -1
		if len(self.args) > 0:
			start = 0
			if self.args[0].name == "this_":
				start = 1

		arguments = ""
		if len(self.args) > 1 or start == 0:
			arguments = ", "
			for i in range(start, len(self.args)):
				arguments += "py::arg(\"{0}\")".format(self.args[i].name, self.args[i].type)

				if self.args[i].hasDefault():
					arguments += " = {0}(".format(self.args[i].type.strip("&").lstrip("const").strip()) + self.args[i].default + ")"

				if i + 1 < len(self.args):
					arguments += ", "

		funcType = None
		if self.static:
			funcType = ".def_static"
		else:
			funcType = ".def"

		return "{4}(\"{0}\", []({1}) {{ {2} }}{3})".format(self.name, inputArgs, self.op, arguments, funcType)



# returns:
#   solution file version
#   visual studio version
#   microsoft compiler version
#   microsoft compiler version (long)
#   vcx project tools version
#   windows SDK version
vs_info = { 'solution':'12',
            'visual studio':'19',
            'msvc':'19',
            'msvc_long':'19.20.27027.1',
            'vcx_tool':'15.0.25914.0',
            'platform_toolset':'142',
            'windows_sdk':'10.0'              # Windows 10 (latest installed)
            # 'windows_sdk':'10.0.17763.0'    # Windows 10 version 1809
            # 'windows_sdk':'10.0.17134.0'    # Windows 10 version 1803
            # 'windows_sdk':'10.0.16299.0'    # Windows 10 Creators Update
            # 'windows_sdk':'10.0.14393.795'  # Windows 10
          }



# generate Visual Studio IDE Filter

from os import makedirs
from os.path import dirname, normpath, join, split, relpath
from errno import EEXIST

def filter_folders(cf_list, af_list, outf):

  f1 = r'''  <ItemGroup>
    <Filter Include="Header Files" />
    <Filter Include="Source Files" />
'''
  f2 = r'''    <Filter Include="Source Files\{0:s}" />
'''
  f3 = r'''  </ItemGroup>
'''
  c_dirs = set(i[2] for i in cf_list)
  a_dirs = set(i[2] for i in af_list)
  if a_dirs:
    c_dirs |= set((r'mpn\yasm',))
  outf.write(f1)
  for d in sorted(c_dirs):
    if d:
      t = d if d != r'mpn\generic' else r'mpn'
      outf.write(f2.format(t))
  outf.write(f3)

def filter_headers(hdr_list, relp, outf):

  f1 = r'''  <ItemGroup>
'''
  f2 = r'''    <ClInclude Include="{}{}">
      <Filter>Header Files</Filter>
    </ClInclude>
'''
  f3 = r'''  </ItemGroup>
'''
  outf.write(f1)
  for h in hdr_list:
    outf.write(f2.format(relp, h))
  outf.write(f3)

def filter_csrc(cf_list, relp, outf):

  f1 = r'''  <ItemGroup>
'''
  f2 = r'''    <ClCompile Include="{}{}">
      <Filter>Source Files</Filter>
    </ClCompile>
'''
  f3 = r'''    <ClCompile Include="{}{}\{}">
      <Filter>Source Files\{}</Filter>
    </ClCompile>
'''
  f4 = r'''  </ItemGroup>
'''
  outf.write(f1)
  for i in cf_list:
    if not i[2]:
      outf.write(f2.format(relp, i[0] + i[1]))
    else:
      t = 'mpn' if i[2].endswith('generic') else i[2]
      outf.write(f3.format(relp, i[2], i[0] + i[1], t))
  outf.write(f4)

def filter_asrc(af_list, relp, outf):

  f1 = r'''  <ItemGroup>
'''
  f2 = r'''    <YASM Include="{0:s}{2:s}\{1:s}">
      <Filter>Source Files\mpn\yasm</Filter>
    </YASM>
'''
  f3 = r'''  </ItemGroup>
'''
  outf.write(f1)
  for i in af_list:
    outf.write(f2.format(relp, i[0] + i[1], i[2], i[2]))
  outf.write(f3)

def gen_filter(path, root_dir, hf_list, cf_list, af_list, tools_ver):

  f1 = r'''<?xml version="1.0" encoding="utf-8"?>
<Project ToolsVersion="{0}" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
'''.format(tools_ver)

  f2 = r'''  <ItemGroup>
    <None Include="..\..\..\gmp-h.in" />
  </ItemGroup>
</Project>
'''

  relp = split(relpath(root_dir, path))[0] + '\\'
  try:
    makedirs(split(path)[0])
  except IOError as e:
    if e.errno != EEXIST:
      raise
    else:
      pass

  with open(path, 'w') as outf:

    outf.write(f1)
    filter_folders(cf_list, af_list, outf)
    if hf_list:
      filter_headers(hf_list, relp, outf)
    filter_csrc(cf_list, relp, outf)
    if af_list:
      filter_asrc(af_list, relp, outf)
    outf.write(f2)



'''
---------------------------------------------------------------------------
Copyright (c) 1998-2008, Brian Gladman, Worcester, UK. All rights reserved.

LICENSE TERMS

The redistribution and use of this software (with or without changes)
is allowed without the payment of fees or royalties provided that:

 1. source code distributions include the above copyright notice, this
    list of conditions and the following disclaimer;

 2. binary distributions include the above copyright notice, this list
    of conditions and the following disclaimer in their documentation;

 3. the name of the copyright holder is not used to endorse products
    built using this software without specific written permission.

DISCLAIMER

This software is provided 'as is' with no explicit or implied warranties
in respect of its properties, including, but not limited to, correctness
and/or fitness for purpose.
---------------------------------------------------------------------------
'''

from __future__ import print_function

import string, re, os, sys, shutil

debug = False
dot_labels = True
remove_comment_lines = True
is_linux = False

# translate 64 bit registers to 8 bit form

r08 = { 'rax' :   'al', 'rbx' :   'bl', 'rcx' :   'cl', 'rdx' :   'dl',
        'rsi' :  'sil', 'rdi' :  'dil', 'rbp' :  'bpl',
         'r8' :  'r8b',  'r9' :  'r9b', 'r10' : 'r10b', 'r11' : 'r11b',
        'r12' : 'r12b', 'r13' : 'r13b', 'r14' : 'r14b', 'r15' : 'r15b' }

# translate 64 bit registers to 32 bit form

r32 = { 'rax' :  'eax', 'rbx' :  'ebx', 'rcx' :  'ecx', 'rdx' :  'edx',
        'rsi' :  'esi', 'rdi' :  'edi', 'rbp' :  'ebp',
         'r8' :  'r8d',  'r9' :  'r9d', 'r10' : 'r10d', 'r11' : 'r11d',
        'r12' : 'r12d', 'r13' : 'r13d', 'r14' : 'r14d', 'r15' : 'r15d' }

# regular expression for registers

r_q = r'(?:r[abcd]x|r[sd]i|r[bsi]p)|(?:r8|r9|r1[0-5])|'  # 64 bit
r_d = r'(?:e[abcd]x|e[sd]i|e[bsi]p|r[89]d|r1[0-5]d)|'    # 32 bit
r_w = r'(?:[abcd]x|[sd]i|[bsi]p|r[89]w|r1[0-5]w)|'       # 16 bit
r_b = r'(?:[abcd]l|[ds]il|[bsi]pl|r[89]b|r1[0-5]b)|'     #  8 bit
r_x = r'(?:x?mm1[0-5]|x?mm\d)|(?:mmx1[0-5]|mmx\d)|(?:st\([0-7]\))'

p_r1 = r'(?:\s*%(?P<reg1>' + r_b + r_w + r_d + r_q + r_x + r'))'
p_r2 = r'(?:\s*%(?P<reg2>' + r_b + r_w + r_d + r_q + r_x + r'))'
p_r3 = r'(?:\s*%(?P<reg3>' + r_b + r_w + r_d + r_q + r_x + r'))'
p_r4 = r'(?:\s*%(?P<reg4>' + r_b + r_w + r_d + r_q + r_x + r'))'

# regular expression for immediate (numeric, not symbolic)

p_im = r'\s+\`?\$\'?(?P<imm>[\+\-]?[0-9]+|0x[a-zA-Z0-9]+)'

# regular expressions for labels

p_l1 = r'(?P<lab1>[a-zA-Z$_][a-zA-Z0-9\$_]*|\.[0-9]+)'
p_l2 = r'L\((?P<lab2>[a-zA-Z0-9\$_]+)\)'
p_l3 = r'(?P<lab3>[a-zA-Z$_][a-zA-Z0-9\$_]*|\.[0-9]+)'
p_l4 = r'L\((?P<lab4>[a-zA-Z0-9\$_]+)\)'

p_lr = r'\s*(?:' + p_l2 + r'|' + p_l1 + r')'
p_la = r'\s*(?:' + p_l2 + r'|' + p_l1 + r')\s*(:)'
p_ri = p_lr + r'\(\%rip\)'
p_34 = r'\s*(?:' + p_l4 + r'|' + p_l3 + r')'
p_jt = r'\s+\.long\s+' + p_lr + r'\s*\-\s*' + p_34

# regular expression for dis(r1, r2, mul) forms

p_di = r'(?:(?:\s*(?P<dis>[-+]?[0-9\+\-]*)(?=\())?' + '|' + p_lr + ')' # numeric displacement
p_mu = r'(?:\s*,\s*(?P<mul>[1248]))?\s*(?=\))'   # multiplier (1, 2, 4, 8)
p_t1 = p_di + r'\s*\(' + p_r1 + r'(?:\s*\)|\s*,' + p_r2 + p_mu + r'\s*\))'

# regular expression for instructions

p_in = r'\s*(?P<ins>[a-zA-Z][a-zA-Z0-9]*)'

p_def = r'define\s*\(\s*\`([a-zA-Z_][a-zA-Z0-9_]*)\'\s*\,\s*\`' + p_r1 + '\'\s*\)'

m_f1 = re.compile(p_in + p_r3 + r'\s*,' + p_t1)
m_f2 = re.compile(p_in + p_t1 + r'\s*,' + p_r3)
m_f3 = re.compile(p_in + p_t1)
m_f4 = re.compile(p_in + p_im + r'\s*,' + p_t1)
m_f5 = re.compile(p_in + p_im + r'\s*,' + p_r1)
m_f6 = re.compile(p_in + p_r1 + r'\s*,' + p_r2)
m_f7 = re.compile(p_in + p_r1)
m_g7 = re.compile(p_in + r'\s+\*' + p_r1)
m_f8 = re.compile(p_in + p_im)
m_f9 = re.compile(p_in + p_lr)
m_fa = re.compile(p_in + '(?:' + p_im + '|' + p_r1 + r')\s*,' + p_r2 + r'\s*,' + p_r3)
m_fb = re.compile(p_in + p_t1 +  r'\s*,' + p_r3 +  r'\s*,' + p_r4)

m_la = re.compile(p_la)
m_jt = re.compile(p_jt)
m_ri = re.compile(p_in + p_ri)

r_mac = re.compile(r"^\s*(?:define|DEFINE)\s*\(\s*`"
                   "([A-Z$_][A-Z0-9$_]*)'\s*,\s*`\s*$")
r_mnd = re.compile(r"^\s*'\s*\)\s*$")
r_dlr = re.compile(r'\$([0-9]+)')
r_def = re.compile(p_def)

r1 = r'\s*(\%{0,1}[0-9]+){0,1}\s*'
r2 = r'(?:,\s*(\%{0,1}[0-9]+))?\s*'
r_mrz = re.compile(r'([A-Z$_][A-Z0-9$_]*)\s*$')
r_mrf = re.compile(r'([A-Z$_][A-Z0-9$_]*)\s*\(' + r1 + r2 +
                  r2 + r2 + r2 + r2 + r2 + r2 + r2 + r2 + r'\)')

def lab_ref(v, labels, macros, mac_name):
  if mac_name and v in macros[mac_name][3]:
    return '%%{0}'.format(macros[mac_name][3][v])
  elif dot_labels:
    return '.{0}'.format(labels.index(v))
  else:
    return 'L_{0}'.format(re.sub('\$', '%', v))

def pass_one(code):
  labels = []
  for l in code:
    m = re.match(r'(\s*)(?:C|#|;|dnl)\s*(.*)', l)
    if m:
      v = list(m.groups())
      continue
    m = m_la.match(l)
    if m:
      lb = m.group(1) if m.group(1) else m.group(2)
      if lb:
        labels += [lb]
  return labels

def pass_two(code, labels):

  mac_name = ''
  mac_dict = {}
  lab_idx = 0
  for ln in range(len(code)):
    l = code[ln]

    m = r_mac.search(l)
    if m and not mac_name:
      mac_name = m.group(1)
      mac_dict[mac_name] = [ln, 0, 0, {}]
      continue

    m = r_mnd.search(l)
    if m and mac_name:
      mac_dict[mac_name][1] = ln
      mac_name = ''
      lab_idx = 0
      continue

    if mac_name:
      m = r_dlr.findall(l)
      if m:
        mn = max(int(i) for i in m)
        if mn > mac_dict[mac_name][2]:
          mac_dict[mac_name][2] = mn

      m = m_la.match(l)
      if m:
        lb = m.group(1) if m.group(1) else m.group(2)
        if lb in labels:
          lab_idx += 1
          mac_dict[mac_name][3][lb] = lab_idx

  return mac_dict

def addr(dd, labels, macros, mac_name):
  pfx = '['
  ss = ']'
  if dd['lab1'] or dd['lab2']:
    lb = dd['lab1'] if dd['lab1'] else dd['lab2']
    if lb in labels:
      ss = lab_ref(lb, labels, macros, mac_name) + ']'
      pfx = '[rel '
      if dd['reg1'] == 'rip':
        return pfx + ss
      ss = '+' + ss
  elif dd['dis']:
    ss = (dd['dis'] if not dd['dis'] or dd['dis'][0] in '+-'
            else '+' + dd['dis']) + ']'
  ss = ('' if not dd['mul'] else '*' + dd['mul']) + ss
  if dd['reg2']:
    ss = '+' + dd['reg2'] + ss
  if dd['reg1']:
    ss = dd['reg1'] + ss
  return pfx + ss

def pass_three(code, labels, macros, level):

  lo = []
  mac_name = ''
  for l in code:

    m = re.match(r'(\s*)(?:C|#|;|dnl)\s*(.*)', l)
    if m:
      v = list(m.groups())
      if not remove_comment_lines:
        lo += ['{0[0]}; {0[1]}'.format(v)]
      continue

    m = re.search(r'(.*)\s+#(.*)$', l)
    sfx = ''
    if m:
      if m.group(1):
        l = m.group(1) + ' '
      else:
        continue
      if m.group(2):
        sfx = '\t;' + m.group(2)

    # labels
    lp = ''
    m = m_la.match(l)
    if m:
      lb = m.group(1) if m.group(1) else m.group(2)
      l = l[ l.find(':') + 1 : -1 ]
      if mac_name:
        if lb in macros[mac_name][3]:
          ii = macros[mac_name][3][lb]
        else:
          print('internal error')
      else:
        ii = labels.index(lb)
      if lb:
        lb = re.sub('\$', '%', lb)
        if not l:
          if mac_name:
            lo += ['\n%%{0}:'.format(ii)]
          elif dot_labels:
            lo += ['\n.{0}:'.format(ii)]
          else:
            lo += ['\nL_{0}:'.format(lb)]
            continue
        else:
          if mac_name:
            lp = '\n%%{0}:'.format(ii)
          elif dot_labels:
            lp = '\n.{0}:'.format(ii)
          else:
            lp = '\nL_{0}:'.format(lb)

    if not re.search(r'(\S+)', l):
      if lp:
        lo += [lp]
        continue
    elif not lp:
      lp = '\n'

    # three operand instructions
    m = m_fa.search(l)
    if m:
      d = m.groupdict()
      if debug:
        print(l, end = '')
      e = d['imm'] if d['imm'] else d['reg1']
      lo += [lp + '\t{0[ins]:7s} {0[reg3]}, {0[reg2]}, {1}'.format(d, e)]
      continue
    
    m = m_fb.search(l)
    if m:
      d = m.groupdict()
      if debug:
        print(l, end = '')
      s = addr(d, labels, macros, mac_name)
      lo += [lp + '\t{0[ins]:7s} {0[reg4]}, {0[reg3]}, {1}'.format(d, s)]
      continue

    # ins reg, dis(reg, reg, off)
    m = m_f1.search(l)
    if m:
      if debug:
        print(l, end = '')
      d = m.groupdict()
      s = addr(d, labels, macros, mac_name)
      lo += [lp + '\t{0[ins]:7s} {1}, {0[reg3]}'.format(d, s)]
      continue

    # ins dis(reg, reg, off), reg
    m = m_f2.search(l)
    if m:
      if debug:
        print(l, end = '')
      d = m.groupdict()
      s = addr(d, labels, macros, mac_name)
      lo += [lp + '\t{0[ins]:7s} {0[reg3]}, {1}'.format(d, s)]
      continue

    # ins dis(reg, reg, off)
    m = m_f3.search(l)
    if m:
      if debug:
        print(l, end = '')
      d = m.groupdict()
      s = addr(d, labels, macros, mac_name)
      lo += [lp + '\t{0[ins]:7s} {1}'.format(d, s)]
      continue

    # ins imm, dis(reg, reg, off)
    m = m_f4.search(l)
    if m:
      if debug:
        print(l, end = '')
      d = m.groupdict()
      s = addr(d, labels, macros, mac_name)
      lo += [lp + '\t{0[ins]:7s} {1}, {0[imm]}'.format(d, s)]
      continue

    # ins imm, reg
    m = m_f5.search(l)
    if m:
      d = m.groupdict()
      if debug:
        print(l, end = '')
      lo += [lp + '\t{0[ins]:7s} {0[reg1]}, {0[imm]}'.format(d)]
      continue

    # ins reg, reg
    m = m_f6.search(l)
    if m:
      d = m.groupdict()
      if debug:
        print(l, end = '')
      lo += [lp + '\t{0[ins]:7s} {0[reg2]}, {0[reg1]}'.format(d)]
      continue

    # ins reg
    m = m_f7.search(l)
    if m:
      d = m.groupdict()
      if debug:
        print(l, end = '')
      lo += [lp + '\t{0[ins]:7s} {0[reg1]}'.format(d)]
      continue

    # ins *reg
    m = m_g7.search(l)
    if m:
      d = m.groupdict()
      if debug:
        print(l, end = '')
      lo += [lp + '\t{0[ins]:7s} {0[reg1]}'.format(d)]
      continue

    # ins imm
    m = m_f8.search(l)
    if m:
      d = m.groupdict()
      if debug:
        print(l, end = '')
      lo += [lp + '\t{0[ins]:7s} {0[imm]}'.format(d)]
      continue

    # jump table
    m = m_jt.search(l)
    if m:
      d = m.groupdict()
      lb0 = d['lab1'] if d['lab1'] else d['lab2']
      lb1 = d['lab3'] if d['lab3'] else d['lab4']
      if lb0 in labels and lb1 in labels:
        if debug:
          print(l, end = '')
        st1 = lab_ref(lb0, labels, macros, mac_name)
        st2 = lab_ref(lb1, labels, macros, mac_name)
        lo += [lp + '\tdd     ' + st1 + ' - ' + st2]
        continue

    # jump label
    m = m_f9.search(l)
    if m:
      d = m.groupdict()
      lb = d['lab1'] if d['lab1'] else d['lab2']
      if lb and lb in labels:
        if debug:
          print(l, end = '')
        ss = lab_ref(lb, labels, macros, mac_name)
        lo += [lp + '\t{0[ins]:7s} {1}'.format(d, ss)]
        continue

    m = re.search(r'\s*\.byte\s+((?:0x|0X)[0-9a-fA-F]+|[0-9]+)\s*', l)
    if m:
      v = list(m.groups())
      lo += [lp + '\tdb      {0[0]}'.format(v)]
      continue

    # macro definitions
    m = r_mac.search(l)
    if m:
      mac_name = m.group(1)
      if mac_name in macros:
        lo += [lp + '%macro ' + mac_name.lower() + ' '
                        + str(macros[mac_name][2])]
        continue

    m = r_mnd.search(l)
    if m and mac_name:
      mac_name = ''
      lab_ofs = 0
      lo += [lp + '%endmacro']
      continue

    m = r_dlr.findall(l)
    if m:
      l = re.sub(r'\$([0-9]+)', r'%\1', l)

    m = re.search('PROLOGUE\(([a-zA-Z$_][a-zA-Z0-9$_]*)\)', l)
    if m:
      if is_linux:
        lo += [lp + '\tGLOBAL_FUNC {0}'.format(m.group(1))]
      else:
        lo += [lp + '\tWIN64_GCC_PROC {0}'.format(m.group(1))]
      continue

    m = re.search('EPILOGUE\(\)', l)
    if m:
      if is_linux:
        lo += [lp + '\tend']
      else:
        lo += [lp + '\tWIN64_GCC_END']
      continue

    # macro calls
    m = r_mrz.search(l)
    if not m:
      m = r_mrf.search(l)
    if m:
      if m.group(1).lower() == 'align':
        fs = '\talign   {0}' if is_linux else '\txalign  {0}'
        lo += [lp + fs.format(m.group(2))]
        continue
      elif m.group(1) in macros:
        lp += '\t{0:7s}'.format(m.group(1).lower())
        i = 2
        while i <= m.lastindex and m.group(i):
          lp += ' {0}'.format(m.group(i))
          i += 1
        lo += [lp]
        continue
      elif m.group(1) and (m.lastindex == 1 or not m.group(2)):
        lo += [lp + '\t{0}'.format(m.group(1))]
        continue

    if mac_name:
      m = re.search(r'\s*([^%]+)%([0-9]+)\s*', l)
      if m and m.lastindex == 2 and int(m.group(2)) <= macros[mac_name][2]:
        lo += [lp + '\t{0}%{1}'.format(m.group(1).lower(),m.group(2))]
        continue

    # ins
    m = re.search(p_in + r'\s+(.*)', l)
    if m:
      v = list(m.groups())
      if debug:
        print(l, end = '')
      if len(v) == 1 or len(v) == 2 and v[1] == '':
        if v[0] in macros:
          lo += [lp + '\t{0}'.format(v[0].lower())]
        else:
          lo += [lp + '\t{0[0]}'.format(v)]
        continue
      elif v[0] == 'C':
        lo += [lp + ';\t{0}'.format(v[1])]
        continue

    m = re.search(r'include\(.+config.m4.+\)', l)
    if m:
      if is_linux:
        lo += [lp + "%include 'yasm_mac.inc'"]
      else:
        lo += [lp + "%include '"
               + ''.join(['..\\'] * level) + "yasm_mac.inc'"]
      continue

    m = re.search(r'\s*(ret\b.*)', l)
    if m:
      lo += [lp + '{0}'.format(l.rstrip(string.whitespace))]
      continue

    m = re.search(r'\s*\.(align\s+[0-9]*)', l)
    if m:
      lo += [lp + '\t{0}'.format(m.group(1))]
      continue
    
    m = re.search(r'\s*(\S+)', l)
    if m:
      if len(l):
        lo += [lp + '{0} ; < not translated >'
               .format(l.rstrip(string.whitespace))]
      else:
        lo += [lp]
    elif lp:
      lo += [lp]

  return lo + ['\n']

def form_path(p):
  n = str.rfind(p, '\\', 0, -1)
  if n != -1:
    if not os.path.exists(p[ : n + 1]):
      form_path(p[ : n + 1])
  if p[-1] == '\\' and not os.path.exists(p):
    os.mkdir(p)

def conv_lines(code, l):
  labels = pass_one(code)
  macros = pass_two(code, labels)
  code = pass_three(code, labels, macros, l)
  return (labels, macros, code)

def conv_file(f_in, f_out, l):
  f = open(f_in, 'r')
  code = f.readlines()
  f.close()
  labels, macros, code = conv_lines(code, l)
  f = open(f_out, 'w')
  f.writelines(code)
  f.close()

def conv_dirs(s, d, l):
  fd = os.listdir(s)
  for f in fd:
    sp = os.path.join(s,f)
    dp = os.path.join(d,f)
    if os.path.isdir(sp):
      conv_dirs(sp, dp, l + 1)
    elif os.path.isfile(sp):
      if sp != dp and os.path.exists(dp):
        continue
      fn = os.path.basename(f)
      x = os.path.splitext(fn)
      if x[1] == '.asm':
        form_path(dp)
        print("translating '{0}'".format(sp))
        f = open(sp, 'r')
        code = f.readlines()
        f.close()
        if sp == dp:
          rp = os.path.join(s, x[0] + ('.as' if is_linux else '.old'))
          os.rename(sp, rp)
        labels, macros, code = conv_lines(code, l)
        f = open(dp, 'w')
        f.writelines(code)
        f.close()
      elif False:
        form_path(dp)
        shutil.copyfile(sp, dp)

if len(sys.argv) == 1:
  cd = os.getcwd()                   # if run in the build.vc10 directory
  if cd.endswith('build.vc10'):
    cd1 = cd + '\\..\\mpn\\x86_64'   # the GAS assembler directory
    cd2 = cd + '\\..\\mpn\\x86_64w'  # the YASM (Windows) assembler directory
  elif cd.endswith('x86_64'):        # if run in the GAS assembler directory
    cd1 = cd
    cd2 = cd + '\\..\\x86_64w'
  elif cd.endswith('x86_64w'):       # if run in the YASM assembler directory
    cd2 = cd
    cd1 = cd + '\\..\\x86_64'
  else:
    print("cannot locate assembler source directory")
    sys.exit(-1)
elif len(sys.argv) == 3:
  cd1 = sys.argv[1]
  cd2 = sys.argv[2]
else:
  print("invalid program location or input")
  sys.exit(-1)

if cd1 and os.path.exists(cd1):
  if os.path.isdir(cd1):
    if os.path.exists(cd2) and os.path.isdir(cd2):
      print("warning: output directory '{0}' already exists".format(cd2))
    if cd1 != cd2:
      conv_dirs(cd1, cd2, 0)         # convert format from GAS to YASM
  elif os.path.isfile(cd1):
    if not os.path.exists(cd2):
      conv_file(cd1, cd2, 0)
else:
  print('{0} is not a file or directory'.format(cd1))



#
# Python script for running GMP tests
#

import os
import shutil
import string
import copy
import subprocess
import code
import sys
import re

cw, f = os.path.split(__file__)
if cw != '':
  os.chdir(cw)

try:
  fn = r'..\output_params.bat'
  f = open(fn)
  par = f.readlines()
  f.close()
  m1 = re.match(r'\(set ldir\=(.*)\)', par[0])
  m2 = re.match(r'\(set libr\=(.*)\)', par[1])
  m3 = re.match(r'\(set plat\=(.*)\)', par[2])
  m4 = re.match(r'\(set conf\=(.*)\)', par[3])
  if m1 and m2 and m3 and m4:
    d = (m1.group(1), m2.group(1), m3.group(1), m4.group(1))
  else:
    raise OSError(2, 'Incorrect file content', fn)
except OSError as e:
  print(e.strerror, e.filename)
  sys.exit(-1)

tdir = '{}\\{}\\'.format(d[2], d[3])
if d[1] == 'dll':
  shutil.copy("{}\\{}mpir.dll".format(d[0], tdir),  tdir)
  xt = 'Dynamic Link Library'
else:
  xt = 'Static Library'
print('Testing MPIR {:s} ({:s}) in <{:s}> Configuration'.format(xt, d[0][:-1], tdir))

dir_list = []
for x in os.walk('mpir-tests'):
  if x[0] == 'mpir-tests': 
    dir_list += x[1]
  else:
    break
  
prj_list = []
for x in dir_list:
  l = os.listdir('mpir-tests\\' + x)
  for f in l :
    y = os.path.splitext(f)
    if y[1] == '.vcxproj' and y[0] != 'add-test-lib':
      prj_list += [y[0]]
prj_list.sort()

exe_list = []
try :
  l = os.listdir(tdir)
except :
  print("Tests have not been built for this configuration")
  os._exit(-1)

for f in l:
  x = os.path.splitext(f)
  if x[1] == '.exe':
    exe_list += [x[0]]
exe_list.sort()
if len(exe_list) == 0:
  print("No executable test files for this configuration")
  os._exit(-1)

build_fail = 0
run_ok = 0
run_fail = 0
for i in prj_list:
  if i in exe_list:
    ef = tdir + '\\' + i + '.exe'
    try:
      prc = subprocess.Popen( ef, stdout = subprocess.PIPE,
        stderr = subprocess.STDOUT, creationflags = 0x08000000 )
    except Exception as str:
      print(i, ': ERROR (', str, ')')
      run_fail += 1
      continue
    output = prc.communicate()[0]
    if prc.returncode:
      print(i, ': ERROR (', prc.returncode, ' )')
      run_fail += 1
    else:
      print(i, ': success')
      run_ok += 1
    if output:
      print('    ', output.decode(), end = '')
  else:
    print("Build failure for {0}".format(i))
    build_fail += 1
print(build_fail + run_ok + run_fail, "tests:")
if build_fail > 0:
  print("\t{0} failed to build".format(build_fail))
if run_ok > 0:
  print("\t{0} ran correctly".format(run_ok))
if run_fail > 0:
  print("\t{0} failed".format(run_fail))
if len(sys.argv) == 1:
  try:
    input(".. completed - press ENTER")
  except:
    pass
else:
  sys.exit(build_fail + run_fail)




# A Python program to run speed and evaluate the performance of MPIR
# routines.
#
# Copyright (c) 2009, Brian Gladman, Worcester, UK.
#
# This file is part of the MPIR Library.  The MPIR Library is free
# software; you can redistribute it and/or modify it under the terms
# of the GNU Lesser General Public License version 2.1 as published
# by the Free Software Foundation.
#
# The MPIR Library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along
# with the MPIR Library; see the file COPYING.LIB.  If not, write to
# the Free Software Foundation, Inc., 51Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from __future__ import print_function
import sys
import os
import shutil
import string
import copy
import code
import math
import platform
from subprocess import Popen, PIPE, STDOUT

if sys.platform.startswith('win'):
  dir = '.\\x64\\release\\'
else :
  dir = './'

ll = [
     '-c -s 10(10)1000 noop',
     '-c -s 10(10)1000 noop_wxs',
     '-c -s 10(10)1000 noop_wxys',
     '-c -s 10(10)1000 mpn_add_n',
     '-c -s 10(10)1000 mpn_sub_n',
     '-c -s 10(10)1000 mpn_addadd_n',
     '-c -s 10(10)1000 mpn_subadd_n',
     '-c -s 10(10)1000 mpn_addsub_n',
     '-c -s 10(10)1000 mpn_karaadd',
     '-c -s 10(10)1000 mpn_karasub',
     '-c -s 10(10)1000 mpn_addmul_1.3333',
     '-c -s 10(10)1000 mpn_submul_1.3333',
     '-c -s 10(10)1000 mpn_submul_2',
     '-c -s 10(10)1000 mpn_mul_1.3333',
     '-c -s 10(10)1000 mpn_mul_1_inplace.3333',
     '-c -s 10(10)1000 mpn_mul_2',
      '-c -s 10(10)1000 mpn_divrem_euclidean_qr_1.3333',
      '-c -s 10(10)1000 mpn_divrem_euclidean_qr_2',
      '-c -s 10(10)1000 mpn_divrem_euclidean_r_1.3333',
      '-c -s 10(10)1000 mpn_divrem_hensel_qr_1.3333',
      '-c -s 10(10)1000 mpn_divrem_hensel_qr_1_1.3333',
      '-c -s 10(10)1000 mpn_divrem_hensel_qr_1_2.3333',
      '-c -s 10(10)1000 mpn_divrem_hensel_r_1.3333',
      '-c -s 10(10)1000 mpn_rsh_divrem_hensel_qr_1.3333',
      '-c -s 10(10)1000 mpn_rsh_divrem_hensel_qr_1_1.3333',
      '-c -s 10(10)1000 mpn_rsh_divrem_hensel_qr_1_2.3333',
      '-c -s 10(10)1000 mpn_divrem_hensel_rsh_qr_1.3333',
      '-c -s 10(10)1000 mpn_divrem_1.3333',
      '-c -s 10(10)1000 mpn_divrem_1f.3333',
      '-c -s 10(10)1000 mpn_mod_1.3333',
      '-c -s 10(10)1000 mpn_mod_1_1',
      '-c -s 10(10)1000 mpn_mod_1_2',
      '-c -s 10(10)1000 mpn_mod_1_3',
      '-c -s 10(10)1000 mpn_mod_1_k.3',
      '-c -s 10(10)1000 mpn_preinv_divrem_1.3333',
      '-c -s 10(10)1000 mpn_preinv_divrem_1f.3333',
      '-c -s 10(10)1000 mpn_preinv_mod_1.3333',
      '-c -s 10(10)1000 mpn_add_err1_n',
      '-c -s 10(10)1000 mpn_sub_err1_n',
      '-c -s 10(10)1000 mpn_inv_divappr_q',
      '-c -s 10(10)1000 mpn_inv_div_qr',
      '-c -s 10(10)1000 mpn_dc_divappr_q',
      '-c -s 10(10)1000 mpn_dc_div_qr_n',
      '-c -s 10(10)1000 mpn_divrem_1_inv.3333',
      '-c -s 10(10)1000 mpn_divrem_1f_div.3333',
      '-c -s 10(10)1000 mpn_divrem_1f_inv.3333',
      '-c -s 10(10)1000 mpn_mod_1_div.3333',
      '-c -s 10(10)1000 mpn_mod_1_inv.3333',
      '-c -s 10(10)1000 mpn_divrem_2',
      '-c -s 10(10)1000 mpn_divrem_2_div',
      '-c -s 10(10)1000 mpn_divrem_2_inv',
      '-c -s 10(10)1000 mpn_divexact_1.3333',
      '-c -s 10(10)1000 mpn_divexact_by3',
      '-c -s 10(10)1000 mpn_divexact_byff',
      '-c -s 10(10)1000 mpn_divexact_byfobm1.3333',
      '-c -s 10(10)1000 mpn_modexact_1_odd.333',
      '-c -s 10(10)1000 mpn_modexact_1c_odd.333',
      '-c -s 10(10)1000 mpn_mod_34lsub1',
      '-c -s 10(10)1000 mpn_dc_tdiv_qr',
      '-c -s 10(10)1000 mpn_lshift.33',
      '-c -s 10(10)1000 mpn_rshift.33',
      '-c -s 10(10)1000 mpn_lshift1',
      '-c -s 10(10)1000 mpn_rshift1',
      '-c -s 10(10)1000 mpn_double',
      '-c -s 10(10)1000 mpn_half',
      '-c -s 10(10)1000 mpn_lshift2',
      '-c -s 10(10)1000 mpn_rshift2',
      '-c -s 10(10)1000 mpn_and_n',
      '-c -s 10(10)1000 mpn_andn_n',
      '-c -s 10(10)1000 mpn_nand_n',
      '-c -s 10(10)1000 mpn_ior_n',
      '-c -s 10(10)1000 mpn_iorn_n',
      '-c -s 10(10)1000 mpn_nior_n',
      '-c -s 10(10)1000 mpn_xor_n',
      '-c -s 10(10)1000 mpn_xnor_n',
      '-c -s 10(10)1000 mpn_com_n',
      '-c -s 10(10)1000 mpn_not',
      '-c -s 10(10)1000 mpn_popcount',
      '-c -s 10(10)1000 mpn_hamdist',

     '-c -s 10(10)1000 MPN_ZERO',
     '-c -s 10(10)1000 MPN_COPY',
     '-c -s 10(10)1000 MPN_COPY_INCR',
     '-c -s 10(10)1000 MPN_COPY_DECR',

     '-c -s 10(10)1000 count_leading_zeros',
     '-c -s 10(10)1000 gmp_allocate_free',
     '-c -s 10(10)1000 malloc_realloc_free',
     '-c -s 10(10)1000 gmp_allocate_reallocate_free',
     '-c -s 10(10)1000 malloc_free',
     '-c -s 10(10)1000 mpn_umul_ppmm',
     '-c -s 10(10)1000 mpz_add',
     '-c -s 10(10)1000 mpz_init_realloc_clear',
     '-c -s 10(10)1000 mpz_init_clear',
     '-c -s 10(10)1000 udiv_qrnnd',
     '-c -s 10(10)1000 udiv_qrnnd_c',
     '-c -s 10(10)1000 udiv_qrnnd_preinv1',
     '-c -s 10(10)1000 udiv_qrnnd_preinv2',
     '-c -s 10(10)1000 umul_ppmm',
     '-c -s 10(10)1000 mpn_popcount',
     '-c -s 10(10)1000 mpn_hamdist',

     ]

lq = [
     '-c -s 10(10)1000 mpn_dc_divrem_n',
     '-c -s 10(10)1000 mpn_dc_divrem_sb',
     '-c -s 10(10)1000 mpn_dc_tdiv_qr',
     '-c -s 10(10)1000 mpn_kara_mul_n',
     '-c -s 10(10)1000 mpn_kara_sqr_n',
     '-c -s 10(10)1000 mpn_mul_basecase',
     '-c -s 1000(500)10000 -t 10 mpn_mul_fft_full',
     '-c -s 10(10)1000 mpn_mul_n',
     '-c -s 10(10)1000 mpn_sqr_basecase',
     '-c -s 10(10)1000 mpn_sqr_n',
     '-c -s 50(10)1000 mpn_toom3_mul_n',
     '-c -s 50(10)1000 mpn_toom3_sqr_n',
     '-c -s 1(5)100 mpz_powm',
     ]

# run an executable and return its error return value and any output
def run_exe(exe, args, inp) :
  al = {'stdin' : PIPE, 'stdout' : PIPE, 'stderr' : STDOUT }
  if sys.platform.startswith('win') :
    al['creationflags'] = 0x08000000
  p = Popen([exe] + args.split(' '), **al)
  res = p.communicate(inp.encode())[0].decode()
  ret = p.poll()
  return (ret, res)

# output a matrix implemented as a dictionary
def mout(m, n) :
  for r in range(n) :
    print('\n{0:3d}'.format(r), end='')
    for c in range(n) :
      print('{0:18.4f}'.format(m[(r,c)]) , end='')
  print

# output a vector
def vout(v) :
  print('   ' , end='')
  for c in v :
    print('{0:18.4f}'.format(c) , end='')
  print()

# In-place LU matrix decomposition. The diagonal
# elements of the upper triangular matrix U are
# all 1 and are not stored. Pivoting is used and
# the matrix is implemented as a dictionary. It
# is only intended for use with small matrices.

def LU_decompose(A, n) :
  p = [0] * n
  for k in range(n) :
    # find pivot
    p[k] = k
    max = math.fabs(A[(k,k)])
    for j in range(k + 1, n) :
      if max < math.fabs(A[(j,k)]) :
        max = math.fabs(A[(j,k)])
        p[k] = j
    # exchange rows if necessary
    if p[k] != k :
      for j in range(n) :
        A[(k,j)], A[(p[k],j)] = A[(p[k],j)], A[(k,j)]

    # exit if matrix is singular
    if A[(k,k)] == 0.0 :
      return None

    # set upper triangular elements
    for j in range(k + 1,n) :
      A[(k,j)] /= A[(k,k)]

    # update remaining part of original matrix
    for i in range(k + 1, n) :
      for j in range(k + 1, n) :
        A[(i,j)] -= A[(i,k)] * A[(k,j)]

  # return pivot array
  return p

# Use the LU decomposition above to solve the matrix
# equation A x = b for x given A and b

def LU_solve(A, p, b) :
  n = len(p)
  x = [0] * n

  # calculate U x = L^-1 b
  for k in range(n) :
    if p[k] != k :
      b[k], b[p[k]] = b[p[k]], b[k]
    x[k] = b[k]
    for i in range(k) :
      x[k] -= x[i] * A[(k,i)]
    x[k] /= A[(k,k)]

  # back substitute for x = U^-1 (L^-1 b)
  for k in reversed(range(n)) :
    if p[k] != k :
      b[k], b[p[k]], b[p[k]], b[k]
    for i in range(k + 1, n) :
      x[k] -= x[i] *  A[(k,i)]
  return x

def lsq_solve(x, y, n) :
  m = {}  # matrix as dictionary
  v = []  # vector as list
  # set up matrix and vectors for least squares
  for i in range(n) :
    v.append(sum(xx ** i * yy for xx, yy in zip(x, y)))
    for j in range(i, n) :
      m[(i,j)] = m[(j,i)] = sum(xx ** (i + j) for xx in x)
  # decompose the matrix into lower and upper triangular
  # matrices
  p = LU_decompose(m, n)
  if p != None :
    return LU_solve(m, p, v)
  else :
    return None

def do_lsq(x, y, lsq_size) :
  # get least squares coefficients
  f = lsq_solve(x, y, lsq_size)

  # now find the standard deviation from the curve
  s = 0
  for i in range(len(x)) :
    t = sum(f[j] * x[i] ** j for j in range(lsq_size))
    s += (y[i] - t) ** 2
  sd =  2 * math.sqrt(s / len(x))

  # now remove 'outliers' - data points outside twice
  # the standard deviation
  sc = 0
  for i in reversed(range(len(x))) :
    t = sum(f[j] * x[i] ** j for j in range(lsq_size))
    if math.fabs(y[i] - t) > sd :
      del x[i]
      del y[i]
      sc += 1

  # if we had to remove more than 10% of measurements
  # declare that the result is not stable
  if 10 * sc > len(x) :
    return None
  else :
    return f

print('Machine:', platform.processor())
print('Running:', platform.platform())
print('SPEED CURVE (l: no of limbs) cycles: c[0] + c[1] * l + c[2] * l^2')
print('ROUTINE                      ', end = '')
print('        c[0]        c[1]        c[2]')
lines = ''
cnt = 0
lsq_size = 4
for args in ll + lq :
  cnt += 1
  # run speed for each routine in the list above
  ret = run_exe(os.path.join(dir, 'speed'), args, '')
  # parse the output to produce limbs[] and times[]
  x = []
  y = []
  lines = ret[1].split('\n')
  for l in lines :
    if len(l) :
      s = l.split()
      try :
        t = [float(i) for i in s]
      except :
        continue
      x += [t[0]]
      y += [t[1]]

# output the name of the routine
  nn = args.split(' ')[-1]
  print('{0:<30s}'.format(nn) , end='')
  if not len(x) :
#   print(ret[1].strip(), '(failed to parse output)')
    print('(failed to parse output)')
    continue

  q = 0 if args in ll else 1
  rep = q
  while rep < 3 :
    rep += 1
    f = do_lsq(x, y, lsq_size)
    if f != None :
      break
  else :
    print('not stable')
    continue
  if args in lq :
    print('{0[0]:11.1f} {0[1]:11.1f} {0[2]:11.1f}'.format(f))
  else :
    print('{0[0]:11.1f} {0[1]:11.1f}'.format(f))




# returns:
#   solution file version
#   visual studio version
#   microsoft compiler version
#   microsoft compiler version (long)
#   vcx project tools version
#   windows SDK version
vs_info = { 'solution':'12',
            'visual studio':'2012',
            'msvc':'11',
            'msvc_long':'11',
            'vcx_tool':'4.0',
            'platform_toolset':'110',
            'windows_sdk':'8.1',
            # 'windows_sdk':'7.1.51106'
          }



# add a project file to the solution

from collections import defaultdict, OrderedDict
from os.path import join, exists, relpath, split
from uuid import uuid4
from re import compile

folder_guid = "{2150E333-8FDC-42A3-9474-1A3956D46DE8}"
vcxproj_guid = "{8BC9CEB8-8B4A-11D0-8D11-00A0C91BC942}"
pyproj_guid =  "{888888A0-9F3D-457C-B088-3A5042F75D52}"

sol_11 = r'''
Microsoft Visual Studio Solution File, Format Version {0}.00
# Visual Studio {1}
'''

sol_12 = r'''VisualStudioVersion = {0}
MinimumVisualStudioVersion = 10.0.40219.1
'''

sol_2 = '''Project("{}") = "{}", "{}", "{}"
EndProject
'''

sol_3 = r'''Global
'''

sol_4 = r'''	GlobalSection(SolutionConfigurationPlatforms) = preSolution
		Debug|Win32 = Debug|Win32
		Debug|x64 = Debug|x64
		Release|Win32 = Release|Win32
		Release|x64 = Release|x64
	EndGlobalSection
'''

sol_5 = r'''	GlobalSection(ProjectConfigurationPlatforms) = postSolution
'''

sol_6 = r'''	GlobalSection(NestedProjects) = preSolution
'''

sol_7 = '''		{} = {}
'''

sol_8 = r'''	EndGlobalSection
'''

sol_9 = r'''	GlobalSection(SolutionProperties) = preSolution
		HideSolutionNode = FALSE
	EndGlobalSection
'''

sol_10 = r'''EndGlobal
'''

s_guid = r'\s*(\{\w{8}-\w{4}-\w{4}-\w{4}-\w{12}\})\s*'
s_name = r'\s*\"([a-zA-Z][-.\\_a-zA-Z0-9]*\s*)\"\s*'
re_guid = compile(r'\s*\"\s*' + s_guid + r'\"\s*')
re_proj = compile(r'Project\s*\(\s*\"' + s_guid + r'\"\)\s*=\s*'
                  + s_name + r'\s*,\s*' + s_name + r'\s*,\s*\"' + s_guid + r'\"')
re_fmap = compile(r'\s*' + s_guid + r'\s*=\s*' + s_guid)
acfg = r'(Debug|Release)\|(WIn32|x64)'
re_acfg = compile(s_guid + r'\.' + acfg + r'\.ActiveCfg\s*=\s*' + acfg)

class msvc_solution(object):

  def __init__(self, path):
    self.soln_path, self.soln_file = split(path)

    self.g2fldr = {}
    self.g2proj = OrderedDict()
    self.gf2gpl = defaultdict(list)
    self.g2acfg = OrderedDict()

    if exists(path):
      lines = open(path).readlines()
      for i, ln in enumerate(lines):
        m = re_proj.search(ln)
        if m:
          if m.group(1) == folder_guid and m.group(2) == m.group(3):
            # folder guid -> folder name
            self.g2fldr[m.group(4)] = m.group(2)
          elif (m.group(1) == vcxproj_guid and m.group(3).endswith('.vcxproj') or
                m.group(1) == pyproj_guid and m.group(3).endswith('.pyproj')):
            # project guid -> proj_type_guid, proj_name, vcx_path
            self.g2proj[m.group(4)] = (m.group(1), m.group(2), m.group(3))

      for i, ln in enumerate(lines):
        m = re_fmap.search(ln)
        if m:
          if m.group(1) in self.g2proj and m.group(2) in self.g2fldr:
            self.gf2gpl[m.group(2)].append(m.group(1))
        m = re_acfg.search(ln)
        if m:
          g = m.group(1)
          if not g in self.g2acfg.keys():
            self.g2acfg[g] = [0, 1, 2, 3]
          ix = 2 * (m.group(2) == 'Release') + (m.group(3) == 'x64')
          ac = 2 * (m.group(4) == 'Release') + (m.group(5) == 'x64')
          self.g2acfg[g][ix] = ac

    for g in self.g2proj:
      for _, gpl in self.gf2gpl.items():
        if g in gpl:
          break
      else:
        self.gf2gpl[''].append(g)

    assert len(self.g2fldr.keys()) == len(self.gf2gpl.keys()) - (1 if '' in self.gf2gpl.keys() else 0)
    assert sum(len(gpl) for gf, gpl in self.gf2gpl.items()) == len(self.g2proj.keys())

  def write_solution(self, vs_info):
    if len(self.g2fldr.keys()) > len(self.gf2gpl.keys()):
      for g in list(self.g2fldr.keys()):
        if g not in self.gf2gpl.keys():
          del self.g2fldr[g]

    assert len(self.g2fldr.keys()) == len(self.gf2gpl.keys()) - (1 if '' in self.gf2gpl.keys() else 0)
    assert sum(len(gpl) for gf, gpl in self.gf2gpl.items()) == len(self.g2proj.keys())

    vs_str = sol_11.format(vs_info['solution'], vs_info['visual studio'])
    if int(vs_info['msvc']) > 12:
      vs_str += sol_12.format(vs_info['msvc_long'])

    with open(join(self.soln_path, self.soln_file), 'w') as outf:
      outf.write(vs_str)

      for g, f in self.g2fldr.items():
        outf.write(sol_2.format(folder_guid, f, f, g))

      for g, (gg, f, n) in self.g2proj.items():
        outf.write(sol_2.format(gg, f, n, g))

      outf.write(sol_3)
      outf.write(sol_4)

      outf.write(sol_5)
      wx, dr = ['Win32', 'x64'], ['Debug', 'Release']
      for g in self.g2proj.keys():
        if g in self.g2acfg.keys():
          ac4 = self.g2acfg.get(g, [0, 1, 2, 3])
          for cf in (0, 1):
            for pl in (0, 1):
              ac = ac4[2 * cf + pl]
              awx = 'AnyCPU' if gg == pyproj_guid else wx[ac & 1]
              adr = dr[(ac > 1) & 1]
              outf.write('		{0:s}.{1:s}|{2:s}.ActiveCfg = {3:s}|{4:s}\n'.format(g, dr[cf], wx[pl], adr, awx))
      outf.write(sol_8)

      del self.gf2gpl['']
      if len(self.gf2gpl.keys()):
        outf.write(sol_6)
        for gf, gpl in self.gf2gpl.items():
          for gp in gpl:
            outf.write(sol_7.format(gp, gf))
        outf.write(sol_8)

      outf.write(sol_9)
      outf.write(sol_10)

  def get_project_guid(self, proj_name, file_path):
    relp = relpath(file_path, self.soln_path)
    p_guid = None
    for g in list(self.g2proj.keys()):
      if self.g2proj[g] == (vcxproj_guid, proj_name, relp):
        p_guid = g
        break
    if not p_guid:
        p_guid = '{' + str(uuid4()).upper() + '}'
    return p_guid

  def set_acfg(self, g, mode):
    if g not in self.g2acfg.keys():
      proj_name = self.g2proj[g][1]
      if len(mode) == 2 or proj_name.endswith('gc') or proj_name.endswith('cxx'):
        cfg = [0, 1, 2, 3]
      else:
        cfg = [0, 0, 2, 2] if mode[0] == 'Win32' else [1, 1, 3, 3]
      self.g2acfg[g] = cfg

  def add_project(self, soln_folder, proj_name, file_path, p_guid, mode):
    relp = relpath(file_path, self.soln_path)
    if soln_folder:
      for g, f in self.g2fldr.items():
        if f == soln_folder:
          f_guid = g
          break
      else:
        f_guid = '{' + str(uuid4()).upper() + '}'
        self.g2fldr[f_guid] = soln_folder

    for g in list(self.g2proj.keys()):
      if self.g2proj[g] == (vcxproj_guid, proj_name, relp):
        break
    else:
      self.g2proj[p_guid.upper()] = (vcxproj_guid, proj_name, relp)
      self.gf2gpl[f_guid if soln_folder else ''].append(p_guid.upper())
      self.set_acfg(p_guid.upper(), mode)



from argumentHelper import *

# Detect LibRapid features
features = []
try:
	with open("../../configuration.txt", "r") as file:
		for line in file:
			if len(line) == 0 or line.startswith("#"):
				continue
			
			print("FEATURE:", line.strip())
			try:
				args = line.split()
				features.append((args[0], args[2])) # Cut out the "="
			except:
				pass
except:
	pass

classDefs = ""
moduleDefs = ""

types = [
	["Complex<float>", "ComplexF32", "float"],
	["Complex<double>", "ComplexF64", "double"],
	["Complex<librapid::mpfr>", "ComplexMPFR", "const librapid::mpfr &"]
]

for type in types:
	t = type[0]
	pretty = type[1]
	scalar = type[2]
	typename = "librapid::{}".format(t)
	constRef = "const librapid::{} &".format(t)
	ref = "librapid::{} &".format(t)
	int64_t = "int64_t"
	extent = "librapid::Extent"
	stdString = "std::string"
	stdStringConstRef = "const std::string &"

	# Class Definition
	classDefs += "py::class_<{0}>(module, \"{1}\")\n".format(typename, pretty)

	# Constructors
	classDefs += "\t.def(py::init<>())\n"
	classDefs += "\t.def(py::init<int64_t>())\n".format(scalar)
	classDefs += "\t.def(py::init<int64_t, int64_t>())\n".format(scalar)

	classDefs += "\t.def(py::init<double>())\n".format(scalar)
	classDefs += "\t.def(py::init<double, double>())\n".format(scalar)

	classDefs += "\t.def(py::init<const lrc::mpfr &>())\n".format(scalar)
	classDefs += "\t.def(py::init<const lrc::mpfr &, const lrc::mpfr &>())\n".format(scalar)

	classDefs += "\t.def(py::init<{}>())\n".format(constRef)

	functions = [
		Function("__add__", [Argument(constRef, "this_"), Argument(constRef, "other")],     "return this_ + other;"),
		Function("__sub__", [Argument(constRef, "this_"), Argument(constRef, "other")],     "return this_ - other;"),
		Function("__mul__", [Argument(constRef, "this_"), Argument(constRef, "other")],     "return this_ * other;"),
		Function("__truediv__", [Argument(constRef, "this_"), Argument(constRef, "other")], "return this_ / other;"),

		Function("__radd__", [Argument(constRef, "this_"), Argument(constRef, "other")],     "return other + this_;"),
		Function("__rsub__", [Argument(constRef, "this_"), Argument(constRef, "other")],     "return other - this_;"),
		Function("__rmul__", [Argument(constRef, "this_"), Argument(constRef, "other")],     "return other * this_;"),
		Function("__rtruediv__", [Argument(constRef, "this_"), Argument(constRef, "other")], "return other / this_;"),

		Function("__iadd__", [Argument(ref, "this_"), Argument(constRef, "other")],     "this_ += other; return this_;"),
		Function("__isub__", [Argument(ref, "this_"), Argument(constRef, "other")],     "this_ -= other; return this_;"),
		Function("__imul__", [Argument(ref, "this_"), Argument(constRef, "other")],     "this_ *= other; return this_;"),
		Function("__itruediv__", [Argument(ref, "this_"), Argument(constRef, "other")], "this_ /= other; return this_;"),
		
		Function("str", [Argument(constRef, "this_"), Argument("int64_t", "base", "10")], "return librapid::str(this_, {-1, base, false});"),
		Function("__str__", [Argument(constRef, "this_")], "return librapid::str(this_, {-1, 10, false});"),
		Function("__repr__", [Argument(constRef, "this_")], "return \"librapid::{}(\\\"\" + librapid::str(this_, {{-1, 10, false}}) + \"\\\")\";".format(pretty)),
	]

	for i in range(len(functions)):
		function = functions[i]
		if isinstance(function,Function):
			classDefs += "\t" + function.gen()
		else:
			classDefs += "\t" + function
		
		if i + 1 < len(functions):
			classDefs += "\n"
		else:
			classDefs += ";\n\n"

functions = []

for type in types:
	t = type[0]
	pretty = type[1]
	scalar = type[2]
	typename = "librapid::{}".format(t)
	constRef = "const librapid::{} &".format(t)
	ref = "librapid::{} &".format(t)
	int64_t = "int64_t"
	extent = "librapid::Extent"
	stdString = "std::string"
	stdStringConstRef = "const std::string &"

	if type not in ["librapid::mpz", "librapid::mpq", "librapid::mpf"]:
		functions += [
			Function("abs", [Argument(constRef, "val")], "return librapid::abs(val);"),
			Function("pow", [Argument(scalar, "base"), Argument(constRef, "power")], "return librapid::pow(base, power);"),
			Function("pow", [Argument(constRef, "base"), Argument(scalar, "power")], "return librapid::pow(base, power);"),
			Function("pow", [Argument(constRef, "base"), Argument(constRef, "power")], "return librapid::pow(base, power);"),
			Function("sqrt", [Argument(constRef, "val")], "return librapid::sqrt(val);"),
			Function("exp", [Argument(constRef, "val")], "return librapid::exp(val);"),
			Function("exp2", [Argument(constRef, "val")], "return librapid::exp2(val);"),
			Function("exp10", [Argument(constRef, "val")], "return librapid::exp10(val);"),
			Function("log", [Argument(constRef, "val")], "return librapid::log(val);"),
			Function("log2", [Argument(constRef, "val")], "return librapid::log2(val);"),
			Function("log10", [Argument(constRef, "val")], "return librapid::log10(val);"),
			Function("log", [Argument(constRef, "val"), Argument(constRef, "base")], "return librapid::log(val, base);"),
			Function("log", [Argument(constRef, "val"), Argument(scalar, "base")], "return librapid::log(val, base);"),

			Function("sin", [Argument(constRef, "val")], "return librapid::sin(val);"),
			Function("cos", [Argument(constRef, "val")], "return librapid::cos(val);"),
			Function("tan", [Argument(constRef, "val")], "return librapid::tan(val);"),

			Function("asin", [Argument(constRef, "val")], "return librapid::asin(val);"),
			Function("acos", [Argument(constRef, "val")], "return librapid::acos(val);"),
			Function("atan", [Argument(constRef, "val")], "return librapid::atan(val);"),
			Function("atan2", [Argument(constRef, "a"), Argument(constRef, "b")], "return librapid::atan2(a, b);"),

			Function("csc", [Argument(constRef, "val")], "return librapid::csc(val);"),
			Function("sec", [Argument(constRef, "val")], "return librapid::sec(val);"),
			Function("cot", [Argument(constRef, "val")], "return librapid::cot(val);"),

			Function("acsc", [Argument(constRef, "val")], "return librapid::acsc(val);"),
			Function("asec", [Argument(constRef, "val")], "return librapid::asec(val);"),
			Function("acot", [Argument(constRef, "val")], "return librapid::acot(val);"),

			Function("sinh", [Argument(constRef, "val")], "return librapid::sinh(val);"),
			Function("cosh", [Argument(constRef, "val")], "return librapid::cosh(val);"),
			Function("tanh", [Argument(constRef, "val")], "return librapid::tanh(val);"),

			Function("asinh", [Argument(constRef, "val")], "return librapid::asinh(val);"),
			Function("acosh", [Argument(constRef, "val")], "return librapid::acosh(val);"),
			Function("atanh", [Argument(constRef, "val")], "return librapid::atanh(val);"),

			Function("arg", [Argument(constRef, "val")], "return librapid::arg(val);"),
			Function("proj", [Argument(constRef, "val")], "return librapid::proj(val);"),
			Function("norm", [Argument(constRef, "val")], "return librapid::norm(val);"),
			Function("polar", [Argument(scalar, "rho"), Argument(scalar, "theta")], "return librapid::polar(rho, theta);"),
		]

for i in range(len(functions)):
	function = functions[i]
	if isinstance(function,Function):
		moduleDefs += "\tmodule" + function.gen()
	else:
		moduleDefs += "\t" + function
	moduleDefs += ";\n"

def write(path:str):
	with open(path, "w") as file:
		file.write("""
#include <librapid/librapid.hpp>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/functional.h>
#include <functional>
#include <string>

// Just remove these. They're pointless
#ifdef min
#undef min
#endif

#ifdef max
#undef max
#endif

namespace lrc = librapid;
namespace py = pybind11;

""")

		file.write("void init_complex(py::module &module) {\n")
		
		file.write(classDefs)
		file.write("\n\n")
		file.write(moduleDefs)

		file.write("\n}")

if __name__ == "__main__":
	write("../autogen/complexInterface.cpp")
	
	print("\"${{CMAKE_CURRENT_SOURCE_DIR}}/src/librapid/python/autogen/complexInterface.cpp\"".format(type))
	print("void init_complex(py::module &);")
	print("init_complex(module);")
	print()



# -*- coding: utf-8 -*-

import os
import platform
import sys
import importlib

ROOT_FILE = importlib.util.find_spec('librapid').origin
ROOT_DIR = os.path.dirname(ROOT_FILE)
sys.path.append(ROOT_DIR)

try:
    if platform.system() == "Windows":
        # Add a load of paths to the DLL search paths
        os.add_dll_directory(ROOT_DIR)

        if "CUDA_PATH" in os.environ:
            os.environ["CUDA_PATH"]
            os.add_dll_directory(os.path.join(os.environ["CUDA_PATH"], "bin"))
except:
    pass

from python.pythonInterface import *



import _librapid
from . import typeMapping
from . import extentInterface

def test(x):
    print(x * 2)

def isArrayObject(obj):
	try:
		# CUDA support
		return isinstance(obj, (_librapid.ArrayB,
								_librapid.ArrayC,
								_librapid.ArrayF16,
								_librapid.ArrayF32,
								_librapid.ArrayF64,
								_librapid.ArrayI16,
								_librapid.ArrayI32,
								_librapid.ArrayI64,
								_librapid.ArrayMPZ,
								_librapid.ArrayMPQ,
								_librapid.ArrayMPFR,
								_librapid.ArrayCF32,
								_librapid.ArrayCF64,
								_librapid.ArrayCMPFR,
								_librapid.ArrayBG,
								_librapid.ArrayCG,
								_librapid.ArrayF16G,
								_librapid.ArrayF32G,
								_librapid.ArrayF64G,
								_librapid.ArrayI16G,
								_librapid.ArrayI32G,
								_librapid.ArrayI64G))
	except:
		# No CUDA support
		return isinstance(obj, (_librapid.ArrayB,
								_librapid.ArrayC,
								_librapid.ArrayF16,
								_librapid.ArrayF32,
								_librapid.ArrayF64,
								_librapid.ArrayI16,
								_librapid.ArrayI32,
								_librapid.ArrayI64,
								_librapid.ArrayMPZ,
								_librapid.ArrayMPQ,
								_librapid.ArrayMPFR,
								_librapid.ArrayCF32,
								_librapid.ArrayCF64,
								_librapid.ArrayCMPFR))

Extent = extentInterface.Extent

mpz = _librapid.mpz
mpf = _librapid.mpf
mpq = _librapid.mpq
mpfr = _librapid.mpfr

ComplexF32 = _librapid.ComplexF32
ComplexF64 = _librapid.ComplexF64
ComplexMPFR = _librapid.ComplexMPFR

prec = _librapid.prec

class Array:
	def __init__(self, *args, **kwargs):
		self._array = None
		type = None
		device = None
		extent = None

		for key, val in kwargs:
			if key in ("extent", "size", "dimensions"):
				if isinstance(val, Extent):
					extent = val
				elif isinstance(val, (list, tuple)):
					extent = Extent(val)
				else:
					raise ValueError("Extent argument must be a list-like object or Extent")
			elif key in ("type", "dtype", "datatype"):
				if isinstance(val, str):
					type = val
				else:
					raise ValueError("Type argument must be a string")
			elif key in ("device", "location", "accelerator"):
				if isinstance(val, str):
					device = val
				else:
					raise ValueError("Device argument must be a string")

		if len(args) > 0:
			if isArrayObject(args[0]):
				self._array = args[0]
			elif isinstance(args[0], Extent):
				extent = args[0]

				if len(args) > 1:
					type = args[1]
				if len(args) > 2:
					device = args[2]
				if len(args) > 3:
					raise ValueError("Array constructor expects <Extent, [type], [device]>")
			else:
				raise ValueError("Extent of Array is required")

		adjustedType, adjustedDevice = typeMapping.mapType(type, device)

		if adjustedType is None or adjustedDevice is None:
			raise ValueError("Array with type={} and device={} is invalid".format(type, device))

		if self._array is None:
			if adjustedDevice == "CPU":
				if adjustedType == "ArrayB":
					self._array = _librapid.ArrayB(extent)
				elif adjustedType == "ArrayC":
					self._array = _librapid.ArrayC(extent._extent)
				elif adjustedType == "ArrayF16":
					self._array = _librapid.ArrayF16(extent._extent)
				elif adjustedType == "ArrayF32":
					self._array = _librapid.ArrayF32(extent._extent)
				elif adjustedType == "ArrayF64":
					self._array = _librapid.ArrayF64(extent._extent)
				elif adjustedType == "ArrayI16":
					self._array = _librapid.ArrayI16(extent._extent)
				elif adjustedType == "ArrayI32":
					self._array = _librapid.ArrayI32(extent._extent)
				elif adjustedType == "ArrayI64":
					self._array = _librapid.ArrayI64(extent._extent)
				elif adjustedType == "ArrayMPZ":
					self._array = _librapid.ArrayMPZ(extent._extent)
				elif adjustedType == "ArrayMPQ":
					self._array = _librapid.ArrayMPQ(extent._extent)
				elif adjustedType == "ArrayMPFR":
					self._array = _librapid.ArrayMPFR(extent._extent)
				elif adjustedType == "ArrayCF32":
					self._array = _librapid.ArrayCF32(extent._extent)
				elif adjustedType == "ArrayCF64":
					self._array = _librapid.ArrayCF64(extent._extent)
				elif adjustedType == "ArrayCMPFR":
					self._array = _librapid.ArrayCMPFR(extent._extent)
			elif adjustedDevice == "GPU":
				if adjustedType == "ArrayB":
					self._array = _librapid.ArrayBG(extent._extent)
				elif adjustedType == "ArrayC":
					self._array = _librapid.ArrayCG(extent._extent)
				elif adjustedType == "ArrayF16":
					self._array = _librapid.ArrayF16G(extent._extent)
				elif adjustedType == "ArrayF32":
					self._array = _librapid.ArrayF32G(extent._extent)
				elif adjustedType == "ArrayF64":
					self._array = _librapid.ArrayF64G(extent._extent)
				elif adjustedType == "ArrayI16":
					self._array = _librapid.ArrayI16G(extent._extent)
				elif adjustedType == "ArrayI32":
					self._array = _librapid.ArrayI32G(extent._extent)
				elif adjustedType == "ArrayI64":
					self._array = _librapid.ArrayI64G(extent._extent)

		if self._array is None:
			raise ValueError("Unknown or invalid Array type")

	def copy(self):
		return Array(self._array.copy())

	def __getitem__(self, index:int):
		return Array(self._array[index])
	
	def __setitem__(self, index:int, val):
		self._array[index] = val

	def __call__(self, *args):
		return self._array(*args)

	def get(self, *args):
		return self._array.get(*args)

	def set(self, val, *args):
		self._array.set(val, *args)

	def scalar(self, index:int):
		return self._array.scalar(index)

	def move(self, newLoc:str):
		_, adjusted = typeMapping.mapType(None, newLoc)
		if adjusted == "CPU":
			return Array(self._array.move_CPU())
		elif adjusted == "GPU":
			return Array(self._array.move_GPU())
		
		raise ValueError("Location \"{}\" is invalid".format(newLoc))

	def cast(self, newType:str):
		adjustedType, _ = typeMapping.mapType(newType, None)
		if adjustedType == "ArrayB":
			return Array(self._array.cast_ArrayB())
		elif adjustedType == "ArrayC":
			return Array(self._array.cast_ArrayC())
		elif adjustedType == "ArrayF16":
			return Array(self._array.cast_Array16F())
		elif adjustedType == "ArrayF32":
			return Array(self._array.cast_ArrayF32())
		elif adjustedType == "ArrayF64":
			return Array(self._array.cast_ArrayF64())
		elif adjustedType == "ArrayI16":
			return Array(self._array.cast_ArrayI16())
		elif adjustedType == "ArrayI32":
			return Array(self._array.cast_ArrayI32())
		elif adjustedType == "ArrayI64":
			return Array(self._array.cast_ArrayI64())
		elif adjustedType == "ArrayMPZ":
			return Array(self._array.cast_ArrayMPZ())
		elif adjustedType == "ArrayMPQ":
			return Array(self._array.cast_ArrayMPQ())
		elif adjustedType == "ArrayMPFR":
			return Array(self._array.cast_ArrayMPFR())
		elif adjustedType == "ArrayCF32":
			return Array(self._array.cast_ArrayCF32())
		elif adjustedType == "ArrayCF64":
			return Array(self._array.cast_ArrayCF64())
		elif adjustedType == "ArrayCMPFR":
			return Array(self._array.cast_ArrayCMPFR())

	def castMove(self, newType:str, newLoc:str):
		adjustedType, adjustedDevice = typeMapping.mapType(newType, newLoc)

		return self.cast(adjustedType).move(adjustedDevice)

	def __add__(self, other):
		if isinstance(other, Array):
			return Array(self._array + other._array)
		return Array(self._array + other)

	def __sub__(self, other):
		if isinstance(other, Array):
			return Array(self._array - other._array)
		return Array(self._array - other)

	def __mul__(self, other):
		if isinstance(other, Array):
			return Array(self._array * other._array)
		return Array(self._array * other)

	def __div__(self, other):
		if isinstance(other, Array):
			return Array(self._array / other._array)
		return Array(self._array / other)

	def __or__(self, other):
		if isinstance(other, Array):
			return Array(self._array | other._array)
		return Array(self._array | other)

	def __and__(self, other):
		if isinstance(other, Array):
			return Array(self._array & other._array)
		return Array(self._array & other)

	def __xor__(self, other):
		if isinstance(other, Array):
			return Array(self._array ^ other._array)
		return Array(self._array ^ other)

	def __neg__(self):
		return Array(-self._array)

	def transpose(self, order:Extent=[]):
		self._array.transpose(order)

	def transposed(self, order:Extent=[]):
		return Array(self._array.transposed(order))

	def dot(self, other):
		return Array(self._array.dot(other._array))

	def str(self, format:str="{}",
				  delim:str=" ",
				  stripWidth:int=-1,
				  beforePoint:int=-1,
				  afterPoint:int=-1,
				  depth:int=0):
		return self._array.str(format,
							  delim,
							  stripWidth,
							  beforePoint,
							  afterPoint,
							  depth)

	def __str__(self):
		return self.str()

	def __repr__(self):
		return repr(self._array)

	def isScalar(self):
		return self._array.isScalar()

	def extent(self):
		return self._array.extent()

def add(lhs:Array, rhs:Array, dst:Array):
	_librapid.add(lhs._array, rhs._array, dst._array)

def sub(lhs:Array, rhs:Array, dst:Array):
	_librapid.sub(lhs._array, rhs._array, dst._array)

def mul(lhs:Array, rhs:Array, dst:Array):
	_librapid.mul(lhs._array, rhs.a_arrayrray, dst._array)

def div(lhs:Array, rhs:Array, dst:Array):
	_librapid.div(lhs._array, rhs._array, dst._array)

# Utility Functions
setNumThreads = _librapid.setNumThreads
setBlasThreads = _librapid.setBlasThreads

# Functions from the Math library
abs = _librapid.abs
floor = _librapid.floor
ceil = _librapid.ceil
pow = _librapid.pow
sqrt = _librapid.sqrt
exp = _librapid.exp
exp2 = _librapid.exp2
exp10 = _librapid.exp10
log = _librapid.log
log2 = _librapid.log2
log10 = _librapid.log10
sin = _librapid.sin
cos = _librapid.cos
tan = _librapid.tan
asin = _librapid.asin
acos = _librapid.acos
atan = _librapid.atan
csc = _librapid.csc
sec = _librapid.sec
cot = _librapid.cot
acsc = _librapid.acsc
asec = _librapid.asec
acot = _librapid.acot
sinh = _librapid.sinh
cosh = _librapid.cosh
tanh = _librapid.tanh
asinh = _librapid.asinh
acosh = _librapid.acosh
atanh = _librapid.atanh
mod = _librapid.mod
round = _librapid.round
roundSigFig = _librapid.roundSigFig
roundTo = _librapid.roundTo
roundUpTo = _librapid.roundUpTo
map = _librapid.map
random = _librapid.random
randint = _librapid.randint
trueRandomEntropy = _librapid.trueRandomEntropy
trueRandom = _librapid.trueRandom
trueRandint = _librapid.trueRandint
randomGaussian = _librapid.randomGaussian
pow10 = _librapid.pow10
lerp = _librapid.lerp

constPi = _librapid.constPi
constEuler = _librapid.constEuler
constLog2 = _librapid.constLog2
constCatalan = _librapid.constCatalan

norm = _librapid.norm
polar = _librapid.polar

toMpz = _librapid.toMpz
toMpf = _librapid.toMpf
toMpq = _librapid.toMpq
toMpfr = _librapid.toMpfr



'''
Set up Visual Sudio to build a specified MPIR configuration

Copyright (C) 2011, 2012, 2013, 2014 Brian Gladman
'''

from operator import itemgetter
from os import scandir, walk, unlink, makedirs, mkdir
from os.path import join, split, splitext, isdir, isfile, exists
from os.path import dirname, abspath, relpath, realpath
from copy import deepcopy
from sys import argv, exit, path
from filecmp import cmp
from shutil import copy
from re import compile, search
from collections import defaultdict
from uuid import uuid4
from time import sleep

from _msvc_filters import gen_filter
from _msvc_project import Project_Type, gen_vcxproj
from _msvc_solution import msvc_solution

# copy from file ipath to file opath but avoid copying if
# opath exists and is the same as ipath (this is to avoid
# triggering an unecessary rebuild).

def write_f(ipath, opath):
  if exists(ipath) and not isdir(ipath):
    if exists(opath) and isfile(opath) and cmp(ipath, opath):
      return
    dp, f = split(opath)
    try:
      mkdir(dp)
    except FileExistsError:
      pass
    copy(ipath, opath)

vs_version = 22
if len(argv) > 1:
  vs_version = int(argv[1])

solution_name = 'mpir.sln'
build_dir_name = 'vs{0:d}'.format(vs_version)

build_root_dir = dirname(realpath(__file__))
mpir_root_dir, build_root_name  = split(build_root_dir)
solution_dir = join(build_root_dir, build_dir_name)
path.append(solution_dir)

write_f(join(build_root_dir, f'version_info{vs_version}.py'),
                            join(solution_dir, 'version_info.py'))
from version_info import vs_info

if len(argv) > 2:
  vs_info['windows_sdk'] = argv[2]

# for script debugging
debug = False
# either add a prebuild step to the project files or do it here
add_prebuild = True
# output a build project for the C++ static library
add_cpp_lib = True

# The path to the mpir root directory
cf_dir = './'
cfg_dir = join(solution_dir, 'cdata')

# paths that might include source files(*.c, *.h, *.asm)
c_directories = ('', build_root_name, 'fft', 'mpf', 'mpq', 'mpz', 'printf', 'scanf')

# files that are to be excluded from the build
exclude_file_list = ('config.guess', 'cfg', 'getopt', 'getrusage',
                     'gettimeofday', 'cpuid', 'obsolete', 'win_timing',
                     'gmp-mparam', 'tal-debug', 'tal-notreent', 'new_fft',
                     'new_fft_with_flint', 'compat', 'udiv_w_sdiv')

# append a file (ipath) to an existing file (opath)

def append_f(ipath, opath):
  try:
    with open(opath, 'ab') as out_file:
      try:
        with open(ipath, 'rb') as in_file:
          buf = in_file.read(8192)
          while buf:
            out_file.write(buf)
            buf = in_file.read(8192)
      except IOError:
        print('error reading {0:s} for input'.format(f))
        return
  except IOError:
    print('error opening {0:s} for output'.format(opath))

# copy files in a list from in_dir to out_dir
def copy_files(file_list, in_dir, out_dir):
  try:
    makedirs(out_dir)
  except IOError:
    pass
  for f in file_list:
    copy(join(in_dir, f), out_dir)

# Recursively search a given directory tree to find header,
# C and assembler code files that either replace or augment
# the generic C source files in the input list 'src_list'.
# As the directory tree is searched, files in each directory
# become the source code files for the current directory and
# the default source code files for its child directories.
#
# Lists of default header, C and assembler source code files
# are maintained as the tree is traversed and if a file in
# the current directory matches the name of a file in the
# default file list (name matches ignore file extensions),
# the name in the list is removed and is replaced by the new
# file found. On return each directory in the tree had an
# entry in the returned dictionary that contains:
#
# 1. The list of header files
#
# 2. The list of C source code files for the directory
#
# 3. The list of assembler code files that replace C files
#
# 4. The list of assembler files that are not C replacements
#

def find_asm(path, cf_list):
  d = dict()
  for root, dirs, files in walk(path):
    # ignore repository directories and the fat directory
    for rd in ('.svn', '.git', '.vs', 'fat'):
      if rd in dirs:
        dirs.remove(rd)
    relp = relpath(root, path)          # path from asm root
    relr = relpath(root, mpir_root_dir) # path from MPIR root
    if relp == '.':                     # set C files as default
      relp = h = t = ''
      d[''] = [[], deepcopy(cf_list), [], [], relr]
    else:
      h, _ = split(relp)                # h = parent, t = this directory
      # copy defaults from this directories parent
      d[relp] = [deepcopy(d[h][0]), deepcopy(d[h][1]),
                 deepcopy(d[h][2]), deepcopy(d[h][3]), relr]
    for f in files:                     # for the files in this directory
      n, x = splitext(f)
      if x == '.h':                     # if it is a header file, remove
        for cf in reversed(d[relp][0]): # any matching default
          if cf[0] == n:
            d[relp][0].remove(cf)
        d[relp][0] += [(n, x, relr)]    # and add the local header file
      if x == '.c':                     # if it is a C file, remove
        for cf in reversed(d[relp][1]): # any matching default
          if cf[0] == n:
            d[relp][1].remove(cf)
        d[relp][1] += [(n, x, relr)]    # and add the local C file
      if x == '.asm':                   # if it is an assembler file
        match = False
        for cf in reversed(d[relp][1]): # remove any matching C file
          if cf[0] == n:
            d[relp][1].remove(cf)
            match = True
            break
        for cf in reversed(d[relp][2]): # and remove any matching
          if cf[0] == n:                # assembler file
            d[relp][2].remove(cf)
            match = True
            break
        if match:                       # if a match was found, put the
          d[relp][2] += [(n, x, relr)]  # file in the replacement list
        else:                           # otherwise look for it in the
          for cf in reversed(d[relp][3]):  # additional files list
            if cf[0] == n:
              d[relp][3].remove(cf)
              break
          d[relp][3] += [(n, x, relr)]
  for k in d:                           # additional assembler list
    for i in range(4):
      d[k][i].sort(key=itemgetter(0))   # sort the four file lists
  return d

# create 4 lists of c, h, cc (or cpp) and asm (or as) files in a directory

def find_src(dir_list):
  # list number from file extension
  di = {'.h': 0, '.c': 1, '.cc': 2, '.cpp': 2, '.asm': 3, '.as': 3}
  list = [[], [], [], []]
  for d in dir_list:
    for f in scandir(join(mpir_root_dir, d)):
      if f.name in {'.svn', '.git', '.vs'}:
        continue                        # ignore repository directories
      if f.is_file():
        n, x = splitext(f.name)         # split into name + extension
        if x in di and not n in exclude_file_list:
          list[di[x]] += [(n, x, d)]    # if of the right type and is
  for x in list:                        # not in the exclude list
    x.sort(key=itemgetter(2, 0, 1))     # add it to appropriate list
  return list

# scan the files in the input set and find the symbols
# defined in the files
fr_sym = compile(r'LEAF_PROC\s+(\w+)')
lf_sym = compile(r'FRAME_PROC\s+(\w+)')
wf_sym = compile(r'WIN64_GCC_PROC\s+(\w+)')
g3_sym = compile(r'global\s+___g(\w+)')
g2_sym = compile(r'global\s+__g(\w+)')

def get_symbols(setf, sym_dir):
  for f in setf:
    fn = join(mpir_root_dir, f[2], f[0] + f[1])
    with open(fn, 'r') as inf:
      lines = inf.readlines()
      for l in lines:
        m = fr_sym.search(l)
        if m:
          sym_dir[f] |= set((m.groups(1)[0],))
        m = lf_sym.search(l)
        if m:
          sym_dir[f] |= set((m.groups(1)[0],))
        m = wf_sym.search(l)
        if m:
          sym_dir[f] |= set((m.groups(1)[0],))
        m = g3_sym.search(l)
        if m:
          sym_dir[f] |= set((m.groups(1)[0],))
        else:
          m = g2_sym.search(l)
          if m:
            sym_dir[f] |= set((m.groups(1)[0],))

def file_symbols(cf):
  sym_dir = defaultdict(set)
  for c in cf:
    if c == 'fat':
      continue
    setf = set()
    for f in cf[c][2] + cf[c][3]:
      setf |= set((f,))
    get_symbols(setf, sym_dir)
  return sym_dir

def gen_have_list(c, sym_dir, out_dir):
  set_sym2 = set()
  for f in c[2]:
    set_sym2 |= sym_dir[f]
  set_sym3 = set()
  for f in c[3]:
    set_sym3 |= sym_dir[f]
  c += [sorted(list(set_sym2)), sorted(list(set_sym3))]
  fd = join(out_dir, c[4])
  try:
    makedirs(fd)
  except IOError:
    pass
  with open(join(fd, 'cfg.h'), 'w') as outf:
    for sym in sorted(set_sym2 | set_sym3):
      print(sym, file=outf)
#   print('/* assembler symbols also available in C files */', file=outf)
#   for sym in sorted(set_sym2):
#     print(sym, file=outf)
#   print('/* assembler symbols not available in C files */', file=outf)
#   for sym in sorted(set_sym3):
#     print(sym, file=outf)

# compile list of C files
t = find_src(c_directories)
c_hdr_list = t[0]
c_src_list = t[1]
if t[2] or t[3]:
  print('found C++ and/or assembler file(s) in a C directory')
  if t[2]:
    for f in t[2]:
      print(f)
    print()
  if t[3]:
    for f in t[3]:
      print(f)
    print()

# compile list of C++ files
t = find_src(['cxx'])
cc_hdr_list = t[0]
cc_src_list = t[2]
if t[1] or t[3]:
  print('found C and/or assembler file(s) in a C++ directory')
  if t[1]:
    for f in t[1]:
      print(f)
    print()
  if t[3]:
    for f in cc_src_list:
      print(f)
    print()

# compile list of C files in mpn\generic
t = find_src([r'mpn\generic'])
gc_hdr_list = t[0]
gc_src_list = t[1]
if t[2] or t[3]:
  print('found C++ and/or assembler file(s) in a C directory')
  if t[2]:
    for f in gc_hdr_list:
      print(f)
    print()
  if t[3]:
    for f in gc_src_list:
      print(f)
    print()

# prepare the generic C build
mpn_gc = dict((('gc', [gc_hdr_list, gc_src_list, [], []]),))

# prepare the list of Win32 builds
mpn_32 = find_asm(join(mpir_root_dir, 'mpn/x86w'), gc_src_list)
syms32 = file_symbols(mpn_32)
del mpn_32['']

# prepare the list of x64 builds
mpn_64 = find_asm(join(mpir_root_dir, 'mpn/x86_64w'), gc_src_list)
syms64 = file_symbols(mpn_64)
del mpn_64['']

nd_gc = len(mpn_gc)
nd_32 = nd_gc + len(mpn_32)
nd_nd = nd_32 + len(mpn_64)

# now ask user which builds they wish to generate

while True:
  cnt = 0
  for v in sorted(mpn_gc):
    cnt += 1
    print('{0:2d}. {1:24s}        '.format(cnt, v.replace('\\', '_')))
  for v in sorted(mpn_32):
    cnt += 1
    print('{0:2d}. {1:24s} (win32)'.format(cnt, v.replace('\\', '_')))
  for v in sorted(mpn_64):
    cnt += 1
    print('{0:2d}. {1:24s}   (x64)'.format(cnt, v.replace('\\', '_')))
  fs = 'Space separated list of builds (1..{0:d}, 0 to exit)? '
  s = input(fs.format(cnt))
  n_list = [int(c) for c in s.split()]
  if 0 in n_list:
    exit()
  if any(n < 1 or n > nd_nd for n in n_list):
    print('list contains invalid build numbers')
    sleep(2)
  else:
    break

# multiple builds must each have their own prebuilds
if len(n_list) > 1:
  add_prebuild = True

# now generate the requested builds

# input any existing projects in the solution (*.sln) file
solc = msvc_solution(abspath(join(solution_dir, solution_name)))
hf_list = ('config.h', 'gmp-impl.h', 'longlong.h', 'mpir.h', 'gmp-mparam.h')
for n in n_list:

  if 0 < n <= nd_gc:
    config = sorted(mpn_gc)[n - 1]
    mode = ('Win32', 'x64')
    mpn_f = mpn_gc[config]
  elif nd_gc < n <= nd_32:
    config = sorted(mpn_32)[n - 1 - nd_gc]
    gen_have_list(mpn_32[config], syms32, cfg_dir)
    mode = ('Win32', )
    mpn_f = mpn_32[config]
  elif nd_32 < n <= nd_nd:
    config = sorted(mpn_64)[n - 1 - nd_32]
    gen_have_list(mpn_64[config], syms64, cfg_dir)
    mode = ('x64', )
    mpn_f = mpn_64[config]
  else:
    print('internal error')
    exit()

  if mode[0] == 'x64':
    for l in mpn_f[1:]:
      for t in l:
        if t[0].startswith('preinv_'):
          if 'x64' in mode and t[0] == 'preinv_divrem_1':
            l.remove(t)

  print(config, mode)

  if not add_prebuild:
    # generate mpir.h and gmp.h from gmp_h.in
    gmp_h = '''
    #ifdef _WIN32
    #  ifdef _WIN64
    #    define _LONG_LONG_LIMB  1
    #    define GMP_LIMB_BITS   64
    #  else
    #    define GMP_LIMB_BITS   32
    #  endif
    #  define __GMP_BITS_PER_MP_LIMB  GMP_LIMB_BITS
    #  define SIZEOF_MP_LIMB_T (GMP_LIMB_BITS >> 3)
    #  define GMP_NAIL_BITS                       0
    #endif
    '''

    try:
      lines = open(join(mpir_root_dir, 'gmp-h.in'), 'r').readlines()
    except IOError:
      print('error attempting to read from gmp_h.in')
      exit()
    try:
      tfile = join(mpir_root_dir, 'tmp.h')
      with open(tfile, 'w') as outf:
        first = True
        for line in lines:
          if search(r'@\w+@', line):
            if first:
              first = False
              outf.writelines(gmp_h)
          else:
            outf.writelines([line])

      # write result to mpir.h but only overwrite the existing
      # version if this version is different (don't trigger an
      # unnecessary rebuild)
      write_f(tfile, join(mpir_root_dir, 'mpir.h'))
      write_f(tfile, join(mpir_root_dir, 'gmp.h'))
      unlink(tfile)
    except IOError:
      print('error attempting to create mpir.h from gmp-h.in')
      exit()

    # generate config.h

    try:
      tfile = join(mpir_root_dir, 'tmp.h')

      if 5 < len(mpn_f) < 8:
        if len(mpn_f) == 6:
          t = sorted(mpn_f[5])
        else:
          t = sorted(mpn_f[5] + mpn_f[6])
        with open(tfile, 'w') as outf:
          for i in t:
            outf.writelines(['#define HAVE_NATIVE_{0:s} 1\n'.format(i)])

      append_f(join(cf_dir, 'cfg.h'), tfile)
      write_f(tfile, join(mpir_root_dir, 'config.h'))
      unlink(tfile)
    except IOError:
      print('error attempting to write to {0:s}'.format(tfile))
      exit()

    # generate longlong.h and copy gmp-mparam.h

    try:
      li_file = None
      for i in mpn_f[0]:
        if i[0] == 'longlong_inc':
          li_file = join(mpir_root_dir, join(i[2], r'longlong_inc.h'))
        if i[0] == 'gmp-mparam':
          write_f(join(mpir_root_dir, join(i[2], 'gmp-mparam.h')),
                  join(mpir_root_dir, 'gmp-mparam.h'))

      if not li_file or not exists(li_file):
        print('error attempting to read {0:s}'.format(li_file))
        exit()

      tfile = join(mpir_root_dir, 'tmp.h')
      write_f(join(mpir_root_dir, 'longlong_pre.h'), tfile)
      append_f(li_file, tfile)
      append_f(join(mpir_root_dir, 'longlong_post.h'), tfile)
      write_f(tfile, join(mpir_root_dir, 'longlong.h'))
      unlink(tfile)
    except IOError:
      print('error attempting to generate longlong.h')
      exit()

  # generate the vcxproj and the IDE filter files
  # and add/replace project in the solution file

  af_list = sorted(mpn_f[2] + mpn_f[3])

  # find the gmp-mparam.h file to be used
  for name, ty, loc in mpn_f[0]:
    if name == 'gmp-mparam':
      loc = loc.replace('mpn\\x86w', '', 1)
      loc = loc.replace('mpn\\x86_64w', '', 1)
      if loc.startswith('\\'):
        loc = loc[1:]
      mp_dir = loc if loc else config
      break
  else:
    mp_dir = config

  cf = config.replace('\\', '_')

  # set up DLL build
  proj_name = 'mpir'
  vcx_name = 'dll_mpir_' + cf
  vcx_path = abspath(join(solution_dir, vcx_name, vcx_name + '.vcxproj'))
  guid = solc.get_project_guid(vcx_name, vcx_path)
  gen_filter(vcx_path + '.filters', mpir_root_dir, hf_list, c_src_list + cc_src_list + mpn_f[1], af_list, 12.0)
  gen_vcxproj(vcx_path, mpir_root_dir, proj_name, guid, mp_dir, mode, Project_Type.DLL,
              False, hf_list, c_src_list + cc_src_list + mpn_f[1], af_list, add_prebuild, vs_info)
  solc.add_project('', vcx_name, vcx_path, guid, mode)

  # set up LIB build
  proj_name = 'mpir'
  vcx_name = 'lib_mpir_' + cf
  vcx_path = abspath(join(solution_dir, vcx_name, vcx_name + '.vcxproj'))
  guid = solc.get_project_guid(vcx_name, vcx_path)
  gen_filter(vcx_path + '.filters', mpir_root_dir, hf_list, c_src_list + mpn_f[1], af_list, 12.0)
  gen_vcxproj(vcx_path, mpir_root_dir, proj_name, guid, mp_dir, mode, Project_Type.LIB,
              False, hf_list, c_src_list + mpn_f[1], af_list, add_prebuild, vs_info)
  solc.add_project('', vcx_name, vcx_path, guid, mode)

# C++ library build

if add_cpp_lib:
  proj_name = 'mpirxx'
  mode = ('Win32', 'x64')
  vcx_name = 'lib_mpir_cxx'
  vcx_path = abspath(join(solution_dir, vcx_name, vcx_name + '.vcxproj'))
  th = hf_list +  ('mpirxx.h',)
  guid = solc.get_project_guid(vcx_name, vcx_path)
  gen_filter(vcx_path + '.filters', mpir_root_dir, th, cc_src_list, '', 12.0)
  gen_vcxproj(vcx_path, mpir_root_dir, proj_name, guid, '', mode, Project_Type.LIB,
              True, th, cc_src_list, '', add_prebuild, vs_info)
  solc.add_project('', vcx_name, vcx_path, guid, mode)

solc.write_solution(vs_info)

# the following code is for diagnostic purposes only
if debug:

  for x in sorted(mpn_f[0] + mpn_f[1]):
    print(x)
  print()
  for x in sorted(mpn_f[2] + mpn_f[3]):
    print(x)
  print()

  # mpn_files = dict()
  # mpn_files.update(mpn_32)
  # mpn_files.update(mpn_64)
  for x in mpn_f[config]:
    print(x)
    if False:
      print('1:')
      for y in mpn_files[x][0]:
        print(y)
      print('2:')
      for y in mpn_files[x][1]:
        print(y)
      print('3:')
      for y in mpn_files[x][2]:
        print(y)
      print('4:')
      for y in mpn_files[x][3]:
        print(y)
      print()
    for y in sorted(x[2] + x[3]):
      print(y)
    print()
    print()

if debug:

  mpn_dirs = ('mpn/generic', 'mpn/x86_64w', 'mpn/x86w')

  # compile a list of files in directories in 'dl' under root 'r' with extension 'p'
  def findf(r, dl, p):
    l = []
    for d in dl:
      for root, dirs, files in walk(r + d):
        relp = relpath(root, r)  # path relative to mpir root directory
        if '.svn' in dirs:
          dirs.remove('.svn')            # ignore SVN directories
        if d == '' or root.endswith(build_vc):
          for d in reversed(dirs):       # don't scan build.vc<nn> subdirectories
            dirs.remove(d)
        for f in files:
          if f.endswith(p):
            l += [(tuple(relp.split('\\')), f)]
    return sorted(l)

  hdr_list = findf(mpir_root_dir, c_directories, '.h')
  for x in hdr_list:
    print(x)
  print()

  src_list = findf(mpir_root_dir, c_directories, '.c')
  for x in src_list:
    print(x)
  print()

  cpp_list = findf(mpir_root_dir, ['cpp'], '.cc')
  for x in cpp_list:
    print(x)
  print()

  gnc_list = findf(mpir_root_dir + 'mpn/', ['generic'], '.c')
  for x in gnc_list:
    print(x)
  print()

  w32_list = findf(mpir_root_dir + 'mpn/', ['x86w'], '.asm')
  for x in w32_list:
    print(x)
  print()

  x64_list = findf(mpir_root_dir + 'mpn/', ['x86_64w'], '.asm')
  for x in x64_list:
    print(x)
  print()

  nd = dict([])
  for d, f in gnc_list:
    n, x = splitext(f)
    nd[n] = nd.get(n, []) + [(d, 'c')]
  for d, f in x64_list:
    n, x = splitext(f)
    nd[n] = nd.get(n, []) + [(d, 'asm')]
  for d, f in w32_list:
    n, x = splitext(f)
    nd[n] = nd.get(n, []) + [(d, 'asm')]
  for x in nd:
    print(x, nd[x])


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
import textwrap
import shutil
import regex

# print("VERSION INFORMATION:")
# os.system("pip show sphinx")
# os.system("pip show breathe")
# os.system("pip show exhale")
# os.system("doxygen --version")

# Update the environment file based on the requirements.txt file
# Note: requirements.txt is automatically udpated by dependabot, while
# the environment file is not. This allows the updates to actually take
# effect
template = """
name: docs # doxygen
channels:
  - conda-forge
dependencies:
  - python=3.9
  - doxygen=1.9.1
  - pip
  - pip:
"""

sys.path.insert(0, os.path.abspath("../.."))

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_dirs = [
	"src/librapid/vendor"
]

file_match = regex.compile(".*\..*")

# -- Project information -----------------------------------------------------

project = "librapid"
copyright = "2022, Toby Davis"
author = "Toby Davis"

# The full version, including alpha/beta/rc tags
version_file = open("../../src/librapid/VERSION.hpp", "r")
release = version_file.readlines()[1].split()[2].replace("\"", "")
version_file.close()

# Set the master file
master_doc = "index"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
	# "furo", <= DO NOT comment back in -- it breaks everything
	"sphinx.ext.autodoc",
	"sphinx.ext.napoleon",
	"sphinx.ext.coverage",
	"sphinx.ext.autosectionlabel",
	"sphinx.ext.intersphinx",
	"sphinx.ext.mathjax",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
	"sphinx_panels",
	"breathe",
	"exhale",
	"numpydoc",
]

autosectionlabel_prefix_document = True

# Set up the breathe extension
breathe_projects = {
	"librapid": "./doxygenoutput/xml"
}

breathe_default_project = "librapid"

# Custom inputs to doxygen generator
doxygen_inputs = """

INPUT				 = ./librapid_doc_copy
	   
ENABLE_PREPROCESSING  = YES
MACRO_EXPANSION	   = YES
EXPAND_ONLY_PREDEF	= NO
PREDEFINED			+= LIBRAPID_DOXYGEN_BUILD
PREDEFINED			+= LR_INLINE=
PREDEFINED			+= __restrict=
PREDEFINED			+= LIBRAPID_MAX_DIMS=32
PREDEFINED			+= __host__=
PREDEFINED			+= __device__=
PREDEFINED			+= __global__=

"""

# Set up the exhale extension
exhale_args = {
	# These arguments are required
	"containmentFolder": "./api",
	"rootFileName": "index.rst",
	"rootFileTitle": "LibRapid",
	"doxygenStripFromPath": "..",
	# Suggested optional arguments
	"createTreeView": True,
	# TIP: if using the sphinx-bootstrap-theme, you need
	# "treeViewIsBootstrap": True,
	"exhaleExecutesDoxygen": True,
	"exhaleDoxygenStdin": textwrap.dedent(doxygen_inputs),
	"verboseBuild": True
}

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
	"*pybind*",
	"*version2*",
	"*jitify*",
	"*fmt*",
	"*blas/*"
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "furo"
# html_theme = "pydata_sphinx_theme"
# html_theme = "sphinx_book_theme"

# html_logo = "../../branding/LibRapid small space.png"

html_context = {
	"github_user": "LibRapid",
	"github_repo": "librapid",
	"github_version": "master",
	"doc_path": "docs/source",
}

html_theme_options = {
	# Show a banner at the top of the page
	# "announcement": """
	#	 <a style=\"text-decoration: none; color: white;\" 
	#		href=\"https://www.patreon.com/librapid\">
	#		<img src=\"/en/latest/_static/LR_icon.png\"
	#		width=22.5/> Support LibRapid on Patreon
	#	 </a>
	# """,
	"sidebar_hide_name": True,
	"light_logo": "LibRapid_light.png",
	"dark_logo": "LibRapid_dark.png",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# These folders are copied to the documentation's HTML output
html_static_path = ["_static"]
html_css_files = ['css/s4defs-roles.css']

html_favicon = "/en/latest/_static/LR_icon.png"

rst_prolog = """
.. include:: <s5defs.txt>
"""




# returns:
#   solution file version
#   visual studio version
#   microsoft compiler version
#   microsoft compiler version (long)
#   vcx project tools version
#   windows SDK version
vs_info = { 'solution':'12',
            'visual studio':'14',
            'msvc':'14',
            'msvc_long':'14.0.24720.0',
            'vcx_tool':'14.0',
            'platform_toolset':'140',
            'windows_sdk':'8.1'            
          }



from argumentHelper import *

# Detect LibRapid features
features = []
try:
	with open("../../configuration.txt", "r") as file:
		for line in file:
			if len(line) == 0 or line.startswith("#"):
				continue
			
			print("FEATURE:", line.strip())
			try:
				args = line.split()
				features.append((args[0], args[2])) # Cut out the "="
			except:
				pass
except:
	pass

resStr = ""

t = "Extent"
typename = "librapid::{}".format(t)
constRef = "const librapid::{} &".format(t)
ref = "librapid::{} &".format(t)
int64_t = "int64_t"
extent = "librapid::Extent"
stdString = "std::string"
stdStringConstRef = "const std::string &"

# Class Definition
resStr += "py::class_<{0}>(module, \"{1}\")\n".format(typename, t)

# Constructors
resStr += "\t.def(py::init<>())\n"
resStr += "\t.def(py::init<const std::vector<int64_t> &>())\n"
resStr += "\t.def(py::init<{}>())\n".format(constRef)

functions = [
	Function("zero", [Argument(int64_t, "dims")], "return librapid::Extent::zero(dims);", static=True),
	Function("stride", [Argument(constRef, "this_")], "return this_.stride();"),
	Function("strideAdjusted", [Argument(constRef, "this_")], "return this_.strideAdjusted();"),
	Function("index", [Argument(constRef, "this_"), Argument(constRef, "ind")], "return this_.index(ind);"),
	Function("indexAdjusted", [Argument(constRef, "this_"), Argument(constRef, "ind")], "return this_.indexAdjusted(ind);"),
	Function("indexAdjusted", [Argument(constRef, "this_"), Argument(constRef, "ind")], "return this_.indexAdjusted(ind);"),
	Function("reverseIndex", [Argument(constRef, "this_"), Argument(int64_t, "ind")], "return this_.reverseIndex(ind);"),
	Function("reverseIndex", [Argument(constRef, "this_"), Argument(int64_t, "ind")], "return this_.reverseIndex(ind);"),
	Function("partial", [Argument(constRef, "this_"), Argument(int64_t, "start", "0"), Argument(int64_t, "end", "-1")], "return this_.partial(start, end);"),
	Function("swivelled", [Argument(constRef, "this_"), Argument(constRef, "order")], "return this_.swivelled(order);"),
	Function("swivel", [Argument(ref, "this_"), Argument(constRef, "order")], "this_.swivel(order);"),
	Function("size", [Argument(constRef, "this_")], "return this_.size();"),
	Function("sizeAdjusted", [Argument(constRef, "this_")], "return this_.sizeAdjusted();"),
	Function("__getitem__", [Argument(constRef, "this_"), Argument(int64_t, "index")], "return this_[index];"),
	Function("__setitem__", [Argument(ref, "this_"), Argument(int64_t, "index"), Argument(int64_t, "val")], "this_[index] = val;"),
	Function("adjusted", [Argument(constRef, "this_"), Argument(int64_t, "ind")], "return this_.adjusted(ind);"),
	Function("__eq__", [Argument(constRef, "this_"), Argument(constRef, "other")], "return this_ == other;"),
	Function("str", [Argument(constRef, "this_")], "return this_.str();"),
	Function("__str__", [Argument(constRef, "this_")], "return this_.str();"),
	Function("__repr__", [Argument(constRef, "this_")], "return \"<librapid::\" + this_.str() + \">\";"),
]

for i in range(len(functions)):
	function = functions[i]
	if isinstance(function,Function):
		resStr += "\t" + function.gen()
	else:
		resStr += "\t" + function
	
	if i + 1 < len(functions):
		resStr += "\n"
	else:
		resStr += ";\n"

def write(path:str):
	with open(path, "w") as file:
		file.write(resStr)

if __name__ == "__main__":
	write("../autogen/extentInterface.hpp")



import _librapid
from . import argCastHelper as caster

class Extent:
	def __init__(self, *args):
		if len(args) == 0:
			self._extent = _librapid.Extent()
		else:
			self._extent = caster.cast(args, int, _librapid.Extent)
			if self._extent is None:
				raise ValueError("Invalid constructor arguments. Input must be list-like object or individual values")

	@staticmethod
	def zero(dims:int):
		return _librapid.Extent.zero(dims)

	def stride(self):
		return Extent(self._extent.stride())

	def strideAdjusted(self):
		return Extent(self._extent.strideAdjusted())

	def index(self, *args):
		tmp = caster.cast(args, int, _librapid.Extent, Extent, lambda x: x._extent)
		if tmp is None:
			raise ValueError("{} is invalid for Extent.index(). Input must be Extent, list-like or integer values".format(args))
		return self._extent.index(tmp._extent)

	def indexAdjusted(self, *args):
		tmp = caster.cast(args, int, _librapid.Extent, Extent, lambda x: x._extent)
		if tmp is None:
			raise ValueError("{} is invalid for Extent.indexAdjusted(). Input must be Extent, list-like or integer values".format(args))
		return self._extent.indexAdjusted(tmp._extent)

	def reverseIndex(self, index:int):
		return Extent(self._extent.reverseIndex(index))

	def partial(self, start:int=0, end:int=-1):
		return Extent(self._extent.partial(start, end))

	def swivelled(self, *args):
		tmp = caster.cast(args, int, _librapid.Extent, Extent, lambda x: x._extent)
		if tmp is None:
			return Extent(self._extent.swivelled(tmp))

	def swivel(self, *args):
		tmp = caster.cast(args, int, _librapid.Extent, Extent, lambda x: x._extent)
		if tmp is None:
			self._extent.swivel(tmp)

	def size(self):
		return self._extent.size()

	def sizeAdjusted(self):
		return self._extent.sizeAdjusted()

	def dims(self):
		return self._extent.dims()

	def __getitem__(self, index:int):
		return self._extent[index]

	def __setitem__(self, index:int, val:int):
		self._extent[index] = val

	def adjusted(self, index:int):
		return self._extent.adjusted(index)

	def __eq__(self, *args):
		tmp = caster.cast(args, int, _librapid.Extent, Extent, lambda x: x._extent)
		if tmp is None:
			raise ValueError("{} is invalid for Extent equality check".format(args))
		return self._extent == tmp

	def __neq__(self, other):
		return not (self == other)

	def str(self):
		return self._extent.str()

	def __str__(self):
		return self._extent.str()

	def __repr__(self):
		return "<librapid::" + self._extent.str() + ">"




from os.path import exists

print('Performing tune prebuild ...')

basic = [
  'dc_bdiv_qr.c', 'dc_bdiv_qr_n.c', 'dc_bdiv_q.c', 'dc_divappr_q.c',
  'dc_div_qr.c', 'dc_div_qr_n.c', 'divrem_2.c',  'divrem_euclidean_r_1.c',
  'divrem_hensel_qr_1.c',  'gcd.c', 'gcdext.c', 'get_str.c', 'hgcd.c',
  'hgcd_appr.c', 'hgcd_reduce.c', 'inv_divappr_q.c', 'inv_div_qr.c',
  'matrix22_mul.c', 'mul.c', 'mulhigh_n.c', 'mullow_n.c', 'mulmod_2expm1.c',
  'mulmod_2expp1_basecase.c', 'mul_n.c', 'rootrem.c', 'rsh_divrem_hensel_qr_1.c',
  'sb_divappr_q.c', 'sb_div_qr.c', 'set_str.c', 'tdiv_q.c', 'tdiv_qr.c',
  'toom3_mul.c', 'toom3_mul_n.c', 'toom4_mul.c', 'toom4_mul_n.c',
  'toom8h_mul.c', 'toom8_sqr_n.c'
]

fft_basic = [
  'adjust.c', 'adjust_sqrt2.c', 'butterfly_lshB.c', 'butterfly_rshB.c',
  'combine_bits.c', 'div_2expmod_2expp1.c', 'fermat_to_mpz.c',
  'fft_mfa_trunc_sqrt2.c', 'fft_mfa_trunc_sqrt2_inner.c', 'fft_negacyclic.c',
  'fft_radix2.c', 'fft_trunc.c', 'ifft_negacyclic.c', 'ifft_radix2.c',
  'ifft_trunc.c', 'ifft_trunc_sqrt2.c', 'mul_2expmod_2expp1.c',
  'mul_fft_main.c', 'mul_mfa_trunc_sqrt2.c', 'mulmod_2expp1.c',
  'mul_trunc_sqrt2.c', 'normmod_2expp1.c', 'revbin.c', 'split_bits.c'
]

for n in basic:
  if not exists(n) :
    with open(n, "wb") as f :
      lines = [('#define TUNE_PROGRAM_BUILD 1\r\n').encode()]
      lines += [('#include "..\\..\\..\\mpn\\generic\\' + n + '"\r\n').encode()]
      f.writelines(lines)

for n in fft_basic:
  if not exists(n) :
    with open(n, "wb") as f :
      lines = [('#define TUNE_PROGRAM_BUILD 1\r\n').encode()]
      lines += [('#include "..\\..\\..\\fft\\' + n + '"\r\n').encode()]
      f.writelines(lines)

n = 'divrem_1.c'
if not exists(n) :
  with open(n, "wb") as f :
    lines = [('#define TUNE_PROGRAM_BUILD 1\r\n').encode()]
    lines += [('#define __gmpn_divrem_1  mpn_divrem_1_tune\r\n').encode()]
    lines += [('#include "..\\..\\..\\mpn\\generic\\' + n + '"\r\n').encode()]
    f.writelines(lines)

n = 'mod_1.c'
if not exists(n) :
  with open(n, "wb") as f :
    lines = [('#define TUNE_PROGRAM_BUILD 1\r\n').encode()]
    lines += [('#define __gmpn_mod_1  mpn_mod_1_tune\r\n').encode()]
    lines += [('#include "..\\..\\..\\mpn\\generic\\' + n + '"\r\n').encode()]
    f.writelines(lines)


n = 'sqr_basecase.asm'
if not exists(n) :
  with open(n, "wb") as f :
    lines = [('%define SQR_KARATSUBA_THRESHOLD_OVERRIDE SQR_KARATSUBA_THRESHOLD_MAX\r\n').encode()]
    lines += [('%include "..\\..\\..\\mpn\\x86_64w\\' + n + '"\r\n').encode()]
    f.writelines(lines)

n = 'fac_ui.c'
if not exists(n) :
  with open(n, "wb") as f :
    lines = [('#define TUNE_PROGRAM_BUILD 1\r\n').encode()]
    lines += [('#define __gmpz_fac_ui mpz_fac_ui_tune\r\n').encode()]
    lines += [('#define __gmpz_oddfac_1 mpz_oddfac_1_tune\r\n').encode()]
    lines += [('#include "..\\..\\..\\mpz\\' + 'oddfac_1.c' + '"\r\n').encode()]
    lines += [('#include "..\\..\\..\\mpz\\' + 'fac_ui.c' + '"\r\n').encode()]
    f.writelines(lines)



import itertools


def generate(types):
	res = []
	for dtype in types:
		tmp = f"""

// ====================================================== //
// The code in this file is GENERATED. DO NOT CHANGE IT.  //
// To change this file's contents, please edit and run    //
// "vec_interface_generator.py" in the same directory     //
// ====================================================== //

py::class_<librapid::{dtype[0]}>(module, \"{dtype[0]}\")
	.def(py::init<>())
	.def(py::init<{", ".join([dtype[-2] for _ in range(dtype[-1])])}>(), {dtype[2]})
	.def(py::init<const librapid::{dtype[0]}>())

	.def("__getitem__", [](const librapid::{dtype[0]} &vec, int64_t index) {{ return vec[index]; }})
	.def("__setitem__", [](librapid::{dtype[0]} &vec, int64_t index, int64_t val) {{ vec[index] = val; }})
	.def("__setitem__", [](librapid::{dtype[0]} &vec, int64_t index, double val) {{ vec[index] = val; }})

	.def("__add__", [](const librapid::{dtype[0]} &lhs, int64_t rhs) {{ return lhs + rhs; }})
	.def("__sub__", [](const librapid::{dtype[0]} &lhs, int64_t rhs) {{ return lhs - rhs; }})
	.def("__mul__", [](const librapid::{dtype[0]} &lhs, int64_t rhs) {{ return lhs * rhs; }})
	.def("__truediv__", [](const librapid::{dtype[0]} &lhs, int64_t rhs) {{ return lhs / rhs; }})

	.def("__add__", [](const librapid::{dtype[0]} &lhs, float rhs) {{ return lhs + rhs; }})
	.def("__sub__", [](const librapid::{dtype[0]} &lhs, float rhs) {{ return lhs - rhs; }})
	.def("__mul__", [](const librapid::{dtype[0]} &lhs, float rhs) {{ return lhs * rhs; }})
	.def("__truediv__", [](const librapid::{dtype[0]} &lhs, float rhs) {{ return lhs / rhs; }})

	.def("__add__", [](const librapid::{dtype[0]} &lhs, double rhs) {{ return lhs + rhs; }})
	.def("__sub__", [](const librapid::{dtype[0]} &lhs, double rhs) {{ return lhs - rhs; }})
	.def("__mul__", [](const librapid::{dtype[0]} &lhs, double rhs) {{ return lhs * rhs; }})
	.def("__truediv__", [](const librapid::{dtype[0]} &lhs, double rhs) {{ return lhs / rhs; }})

	.def("__iadd__", [](librapid::{dtype[0]} &lhs, int64_t rhs) {{ lhs += rhs; }})
	.def("__isub__", [](librapid::{dtype[0]} &lhs, int64_t rhs) {{ lhs -= rhs; }})
	.def("__imul__", [](librapid::{dtype[0]} &lhs, int64_t rhs) {{ lhs *= rhs; }})
	.def("__itruediv__", [](librapid::{dtype[0]} &lhs, int64_t rhs) {{ lhs /= rhs; }})

	.def("__iadd__", [](librapid::{dtype[0]} &lhs, float rhs) {{ lhs += rhs; }})
	.def("__isub__", [](librapid::{dtype[0]} &lhs, float rhs) {{ lhs -= rhs; }})
	.def("__imul__", [](librapid::{dtype[0]} &lhs, float rhs) {{ lhs *= rhs; }})
	.def("__itruediv__", [](librapid::{dtype[0]} &lhs, float rhs) {{ lhs /= rhs; }})

	.def("__iadd__", [](librapid::{dtype[0]} &lhs, double rhs) {{ lhs += rhs; }})
	.def("__isub__", [](librapid::{dtype[0]} &lhs, double rhs) {{ lhs -= rhs; }})
	.def("__imul__", [](librapid::{dtype[0]} &lhs, double rhs) {{ lhs *= rhs; }})
	.def("__itruediv__", [](librapid::{dtype[0]} &lhs, double rhs) {{ lhs /= rhs; }})
	
"""

		for dtype2 in types:
			tmp += f"""
	.def("__add__", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs + rhs; }})
	.def("__sub__", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs - rhs; }})
	.def("__mul__", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs * rhs; }})
	.def("__truediv__", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs / rhs; }})

	.def("__iadd__", [](librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ lhs += rhs; }})
	.def("__isub__", [](librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ lhs -= rhs; }})
	.def("__imul__", [](librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ lhs *= rhs; }})
	.def("__itruediv__", [](librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ lhs /= rhs; }})

	.def("dist2", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs.dist2(rhs); }})
	.def("dist", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs.dist(rhs); }})
"""

		tmp += f"""
	.def("mag2", &librapid::{dtype[0]}::mag2)
	.def("mag", &librapid::{dtype[0]}::mag)
	.def("invMag", &librapid::{dtype[0]}::invMag)
"""

		for dtype2 in types:
			if dtype[-1] != dtype2[-1]:
				continue

			tmp += f"""
	.def("dot", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs.dot(rhs); }}, py::arg("other"))
"""

			if dtype[-2] == 3 and dtype2[-2] == 3:
				tmp += f"""
	.def("cross", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs.cross(rhs); }}, py::arg("other"))
"""

		tmp += f"""
	.def("str", &librapid::{dtype[0]}::str)
	.def("__str__", &librapid::{dtype[0]}::str)
	.def("__repr__", [](const librapid::{dtype[0]} &vec) {{ return "{dtype[0]}" + vec.str(); }})
	.def("__len__", [](const librapid::{dtype[0]} &vec) {{ return {dtype[-1]}; }})
	.def_property("x", &librapid::{dtype[0]}::getX, &librapid::{dtype[0]}::setX)
	.def_property("y", &librapid::{dtype[0]}::getY, &librapid::{dtype[0]}::setY)
	.def_property("z", &librapid::{dtype[0]}::getZ, &librapid::{dtype[0]}::setZ)
	.def_property("w", &librapid::{dtype[0]}::getW, &librapid::{dtype[0]}::setW)
"""

		toSwizzle = ["xy", "xyz", "xyzw"]
		for swiz in toSwizzle:
			for perm in itertools.permutations(list(swiz)):
				joined = "".join(perm)
				tmp += f"\t.def(\"{joined}\", &librapid::{dtype[0]}::{joined})\n"
		tmp = tmp[:-1]
		tmp += ";"

		res.append((dtype[0], tmp))
	return res


types = [
	["Vec2i", "int64_t x, int64_t y", "py::arg(\"x\"), py::arg(\"y\") = 0", "int64_t", 2],
	["Vec2f", "float x, float y", "py::arg(\"x\"), py::arg(\"y\") = 0", "float", 2],
	["Vec2d", "double x, double y", "py::arg(\"x\"), py::arg(\"y\") = 0", "double", 2],

	["Vec3i", "int64_t x, int64_t y, int64_t z", "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0", "int64_t", 3],
	["Vec3f", "float x, float y, float z", "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0", "float", 3],
	["Vec3d", "double x, double y, double z", "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0", "double", 3],

	["Vec4i", "int64_t x, int64_t y, int64_t z, int64_t w", "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0, py::arg(\"w\") = 0", "int64_t", 4],
	["Vec4f", "float x, float y, float z, float w", "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0, py::arg(\"w\") = 0", "float", 4],
	["Vec4d", "double x, double y, double z, double w", "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0, py::arg(\"w\") = 0", "double", 4]
]

def write(path:str):
	res = generate(types)
	for tmp in res:
		with open(f"{path}/{tmp[0]}Interface.cpp", "w") as file:
			file.write("""
	#include <librapid/librapid.hpp>
	#include <pybind11/pybind11.h>
	#include <pybind11/stl.h>
	#include <pybind11/functional.h>
	#include <functional>
	#include <string>

	// Just remove these. They're pointless
	#ifdef min
	#undef min
	#endif

	#ifdef max
	#undef max
	#endif

	namespace lrc = librapid;
	namespace py = pybind11;

	""")

			file.write("void init_{}(py::module &module) {{\n".format(tmp[0]))
			
			file.write(tmp[1])

			file.write("\n}")

if __name__ == "__main__":
	write("../autogen")

	res = generate(types)

	for tmp in res:
		print("\"${{CMAKE_CURRENT_SOURCE_DIR}}/src/librapid/python/autogen/{}\"".format(f"../autogen/{tmp[0]}Interface.cpp"))
	print()
	for tmp in res:
		print(f"void init_{tmp[0]}(py::module &);")
	print()
	for tmp in res:
		print(f"init_{tmp[0]}(module);")
	print()




# returns: 
#   solution file version
#   visual studio version
#   microsoft compiler version
#   microsoft compiler version (long)
#   vcx project tools version
#   windows SDK version
vs_info = { 'solution':'12',
            'visual studio':'15',
            'msvc':'15',
            'msvc_long':'15.0',
            'vcx_tool':'15.0.25914.0',
            'platform_toolset':'141',
            'windows_sdk':'10.0.15063.0'      # Windows 10 Creators Update
            # 'windows_sdk':'10.0.14393.795'  # Windows 10
          }



import librapid as lrp

def testArrayPrint():
    """
    Simple array formatting. This tests indexing as well,
    and if this succeeds, it allows us to test other features
    more easily as we know certain fundamental operations are
    working as expected
    """
    
    x = lrp.ArrayF32(lrp.Extent([5]))
    for i in range(5):
        x[i] = i + 1
    assert(x.str() == "[1 2 3 4 5]")




# generate vcxproj file

from os.path import normpath, join, split, relpath
from os import access, F_OK
from enum import IntEnum

class Project_Type(IntEnum):
  APP = 0
  LIB = 1
  DLL = 2

app_ext = ('.exe', '.lib', '.dll')
app_str = ('Application', 'StaticLibrary', 'DynamicLibrary')

def vcx_proj_cfg(plat, outf):

  f1 = r'''  <ItemGroup Label="ProjectConfigurations">
'''
  f2 = r'''    <ProjectConfiguration Include="{1:s}|{0:s}">
      <Configuration>{1:s}</Configuration>
      <Platform>{0:s}</Platform>
    </ProjectConfiguration>
'''
  f3 = r'''  </ItemGroup>
'''
  outf.write(f1)
  for pl in plat:
    for conf in ('Release', 'Debug'):
      outf.write(f2.format(pl, conf))
  outf.write(f3)

def vcx_globals(name, guid, vs_info, outf):

  f1 = r'''  <PropertyGroup Label="Globals">
    <RootNamespace>{0:s}</RootNamespace>
    <Keyword>Win32Proj</Keyword>
    <ProjectGuid>{1:s}</ProjectGuid>
    <WindowsTargetPlatformVersion>{2:s}</WindowsTargetPlatformVersion>
  </PropertyGroup>
'''
  outf.write(f1.format(name, guid, vs_info['windows_sdk']))

def vcx_default_cpp_props(outf):

  f1 = r'''  <Import Project="$(VCTargetsPath)\Microsoft.Cpp.Default.props" />
'''
  outf.write(f1)

def vcx_library_type(plat, proj_type, vs_info, outf):

  f1 = r'''  <PropertyGroup Condition="'$(Configuration)|$(Platform)'=='{1:s}|{0:s}'" Label="Configuration">
    <ConfigurationType>{2:s}</ConfigurationType>
    <UseDebugLibraries>{3:s}</UseDebugLibraries>
    <PlatformToolset>v{4:s}</PlatformToolset>
  </PropertyGroup>
'''
  for pl in plat:
    for conf, bool in (('Release', 'false'), ('Debug', 'true')):
      outf.write(f1.format(pl, conf, app_str[proj_type], bool, vs_info['platform_toolset']))

def vcx_cpp_props(outf):

  f1 = r'''  <Import Project="$(VCTargetsPath)\Microsoft.Cpp.props" />
'''
  outf.write(f1)

def vcx_extensions(outf, dir):

  f1 = r'''  <ImportGroup Label="ExtensionSettings">
    <Import Project="{0:s}vsyasm.props" />
  </ImportGroup>
'''
  outf.write(f1.format(dir))

def vcx_user_props(plat, proj_type, outf):

  f1 = r'''  <ImportGroup Condition="'$(Configuration)|$(Platform)'=='{1:s}|{0:s}'" Label="PropertySheets">
    <Import Project="$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props" Condition="exists('$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props')" />
    <Import Project="..\..\mpir_{2:s}_{3:s}.props" />
  </ImportGroup>
'''
  for pl in plat:
    for conf in ('Release', 'Debug'):
      outf.write(f1.format(pl, conf, conf.lower(), app_ext[proj_type][1:]))

def vcx_target_name_and_dirs(name, plat, proj_type, outf):

  f1 = r'''  <PropertyGroup>
    <_ProjectFileVersion>10.0.21006.1</_ProjectFileVersion>
'''
  f2 = r'''    <TargetName Condition="'$(Configuration)|$(Platform)'=='{1:s}|{0:s}'">{2:s}</TargetName>
'''
  f3 = r'''  </PropertyGroup>
'''
  outf.write(f1)
  for pl in plat:
    for conf in ('Release', 'Debug'):
      outf.write(f2.format(pl, conf, name))
  outf.write(f3)

def yasm_options(plat, proj_type, outf):

  f1 = r'''    <YASM>
      <Defines>{0:s}</Defines>
      <IncludePaths>..\..\..\mpn\x86{1:s}w\</IncludePaths>
      <Debug>true</Debug>
      <ObjectFile>$(IntDir)mpn\%(FileName).obj</ObjectFile>
    </YASM>
'''

  outf.write(f1.format('DLL' if proj_type == Project_Type.DLL else '', '' if plat == 'Win32' else '_64'))

def compiler_options(plat, proj_type, is_debug, is_avx2, outf):

  f1 = r'''    <ClCompile>
      <AdditionalIncludeDirectories>..\..\..\</AdditionalIncludeDirectories>
      <PreprocessorDefinitions>{0:s}%(PreprocessorDefinitions)</PreprocessorDefinitions>
'''
  f2 = r'''      <EnableEnhancedInstructionSet>AdvancedVectorExtensions2</EnableEnhancedInstructionSet>
'''
  f3 = r'''    </ClCompile>
'''
  if proj_type == Project_Type.APP:
    s1 = 'DEBUG;WIN32;_CONSOLE;'
  if proj_type == Project_Type.DLL:
    s1 = 'DEBUG;WIN32;HAVE_CONFIG_H;MSC_BUILD_DLL;'
  elif proj_type == Project_Type.LIB:
    s1 = 'DEBUG;WIN32;_LIB;HAVE_CONFIG_H;'
  else:
    pass
  if plat == 'x64':
    s1 = s1 + '_WIN64;'
  s1 = ('_' if is_debug else 'N') + s1
  outf.write(f1.format(s1))
  if is_avx2:
    outf.write(f2)
  outf.write(f3)

def linker_options(outf):

  f1 = r'''    <Link>
    </Link>
'''
  outf.write(f1)

def vcx_pre_build(name, plat, vs_info, outf):

  f1 = r'''    <PreBuildEvent>
      <Command>cd ..\..\
prebuild {0:s} {1:s} {2:s}
      </Command>
    </PreBuildEvent>
'''
  outf.write(f1.format(name, plat, vs_info['visual studio']))

def vcx_post_build(is_cpp, vs_info, outf):

  f1 = r'''    <PostBuildEvent>
      <Command>cd ..\..\
postbuild "$(TargetPath)" {0:s}
      </Command>
    </PostBuildEvent>
'''

  outf.write(f1.format(vs_info['visual studio']))

def vcx_tool_options(config, plat, proj_type, is_cpp, af_list, add_prebuild, vs_info, outf):

  f1 = r'''  <ItemDefinitionGroup Condition="'$(Configuration)|$(Platform)'=='{1:s}|{0:s}'">
'''
  f2 = r'''  </ItemDefinitionGroup>
'''
  for pl in plat:
    for is_debug in (False, True):
      outf.write(f1.format(pl, 'Debug' if is_debug else 'Release'))
      if add_prebuild and not is_cpp:
        vcx_pre_build(config, pl, vs_info, outf)
      if af_list:
        yasm_options(pl, proj_type, outf)
      compiler_options(pl, proj_type, is_debug, config.endswith('\\avx'), outf)
      if proj_type != Project_Type.LIB:
        linker_options(outf)
      vcx_post_build(is_cpp, vs_info, outf)
      outf.write(f2)

def vcx_external_props(outf):

  f1 = r'''  <ImportGroup>
    <Import Condition="exists('$(MPIR_Props_External)')" Project="$(MPIR_Props_External)" />
  </ImportGroup>
'''
  outf.write(f1)

def vcx_hdr_items(hdr_list, relp, outf):

  f1 = r'''  <ItemGroup>
'''
  f2 = r'''    <ClInclude Include="{}{}" />
'''
  f3 = r'''  </ItemGroup>
'''
  outf.write(f1)
  for i in hdr_list:
    outf.write(f2.format(relp, i))
  outf.write(f3)

def vcx_c_items(cf_list, plat, relp, outf):

  f1 = r'''  <ItemGroup>
'''
  f2 = r'''    <ClCompile Include="{0:s}{1[0]:s}{1[1]:s}" />
'''
  f3 = r'''    <ClCompile Include="{0:s}{1[2]:s}\{1[0]:s}{1[1]:s}" />
'''
  f6 = r'''  </ItemGroup>
'''
  outf.write(f1)
  for nxd in cf_list:
    if nxd[2] == '':
      outf.write(f2.format(relp, nxd))
    else:
      outf.write(f3.format(relp, nxd))

  outf.write(f6)

def vcx_a_items(af_list, relp, outf):

  f1 = r'''  <ItemGroup>
'''
  f2 = r'''    <YASM Include="{0:s}{1[2]:s}\{1[0]:s}{1[1]:s}" />
'''
  f3 = r'''  </ItemGroup>
'''
  outf.write(f1)
  for nxd in af_list:
    outf.write(f2.format(relp, nxd))
  outf.write(f3)

def gen_vcxproj(path, root_dir, proj_name, guid, config, plat, proj_type,
                is_cpp, hf_list, cf_list, af_list, add_prebuild, vs_info):

  f1 = r'''<?xml version="1.0" encoding="utf-8"?>
<Project DefaultTargets="Build" ToolsVersion="{}" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
'''
  f2 = r'''  <PropertyGroup Label="UserMacros" />
'''
  f3 = r'''  <Import Project="$(VCTargetsPath)\Microsoft.Cpp.targets" />
'''
  f4 = r'''  <ImportGroup Label="ExtensionTargets">
    <Import Project="{0:s}vsyasm.targets" />
    </ImportGroup>
'''

  f5 = r'''<ItemGroup>
    <None Include="..\..\..\gmp-h.in" />
    </ItemGroup>
</Project>
'''

  relp = split(relpath(root_dir, path))[0] + '\\'
  vsy_path = split(path)[0] + '\\..\\vsyasm.targets'
  dir = '..\\' if access(vsy_path, F_OK) else '..\\..\\'
  with open(path, 'w') as outf:
    outf.write(f1.format(vs_info['vcx_tool']))
    vcx_proj_cfg(plat, outf)
    vcx_globals(proj_name, guid, vs_info, outf)
    vcx_default_cpp_props(outf)
    vcx_library_type(plat, proj_type, vs_info, outf)
    vcx_cpp_props(outf)
    if af_list:
      vcx_extensions(outf, dir)
    vcx_user_props(plat, proj_type, outf)
    outf.write(f2)
    vcx_target_name_and_dirs(proj_name, plat, proj_type, outf)
    vcx_tool_options(config, plat, proj_type, is_cpp, af_list, add_prebuild, vs_info, outf)
    vcx_external_props(outf)
    if hf_list:
      vcx_hdr_items(hf_list, relp, outf)
    vcx_c_items(cf_list, plat, relp, outf)
    vcx_a_items(af_list, relp, outf)
    outf.write(f3)
    if af_list:
      outf.write(f4.format(dir))
    outf.write(f5)




# NOTE: Here, None represents a default, unintialized value

typeMap = {
	"ArrayB" : ("b", "bool", "boolean"),
	"ArrayC" : ("c", "char", "byte", "character"),
	"ArrayF16" : ("half", "float16", "f16"),
	"ArrayF32" : ("float", "float32", "f32", "default", None),
	"ArrayF64" : ("double", "float64", "f64"),
	"ArrayI16" : ("short", "int16", "i16"),
	"ArrayI32" : ("int", "long", "int32", "i32"),
	"ArrayI64" : ("long long", "int64", "i64"),
	"ArrayMPZ" : ("bigint", "mpz", "mpz_t", "mpz_class", "mpir_int"),
	"ArrayMPQ" : ("bigrational", "mpq", "mpq_t", "mpq_class", "mpir_rational"),
	"ArrayMPFR" : ("bigfloat", "mpfr", "mpreal", "mpfr_float"),
	"ArrayCF32" : ("complex float", "cfloat", "cf32", "c32"),
	"ArrayCF64" : ("compled double", "cdouble", "cf64", "c64"),
	"ArrayCMPFR" : ("complex multiprecision", "complex multiprec", "cmpfr", "cmpf", "cmp")
}

deviceMap = {
	"CPU" : ("cpu", "host", "default", None),
	"GPU" : ("gpu", "cuda", "device")
}

def mapType(type:str = None, device:str = None):
	resType = None
	resDevice = None
	
	for key, val in typeMap.items():
		if type in val:
			resType = key

	for key, val in deviceMap.items():
		if device in val:
			resDevice = key

	return resType, resDevice
